import logging
from typing import List, Union

from spaceone.core.service import (
    BaseService,
    authentication_handler,
    authorization_handler,
    event_handler,
    mutation_handler,
    transaction,
)
from spaceone.core.service.utils import (
    append_keyword_filter,
    append_query_filter,
    convert_model,
)

from spaceone.identity.manager.role_binding_manager import RoleBindingManager
from spaceone.identity.manager.role_manager import RoleManager
from spaceone.identity.manager.user_manager import UserManager
from spaceone.identity.manager.workspace_group_manager import WorkspaceGroupManager
from spaceone.identity.manager.workspace_group_user_manager import (
    WorkspaceGroupUserManager,
)
from spaceone.identity.model.workspace_group.response import (
    WorkspaceGroupResponse,
    WorkspaceGroupsResponse,
)
from spaceone.identity.model.workspace_group_user.request import (
    WorkspaceGroupUserAddRequest,
    WorkspaceGroupUserFindRequest,
    WorkspaceGroupUserGetRequest,
    WorkspaceGroupUserRemoveRequest,
    WorkspaceGroupUserSearchQueryRequest,
    WorkspaceGroupUserStatQueryRequest,
    WorkspaceGroupUserUpdateRoleRequest,
)
from spaceone.identity.model.workspace_group_user.response import (
    WorkspaceGroupUsersSummaryResponse,
)
from spaceone.identity.service.role_binding_service import RoleBindingService
from spaceone.identity.service.workspace_group_service import WorkspaceGroupService

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class WorkspaceGroupUserService(BaseService):
    resource = "WorkspaceGroupUser"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user_mgr = UserManager()
        self.workspace_group_svc = WorkspaceGroupService()
        self.workspace_group_mgr = WorkspaceGroupManager()
        self.workspace_group_user_mgr = WorkspaceGroupUserManager()
        self.role_mgr = RoleManager()
        self.rb_svc = RoleBindingService()
        self.rb_mgr = RoleBindingManager()

    @transaction(permission="identity:WorkspaceGroupUser:write", role_types=["USER"])
    @convert_model
    def add(
        self, params: WorkspaceGroupUserAddRequest
    ) -> Union[WorkspaceGroupResponse, dict]:
        """Add workspace group user
        Args:
            params (WorkspaceGroupUserAddRequest): {
                'workspace_group_id': 'str',     # required
                'users': [
                    {
                        'user_id': 'str',        # required
                        'role_id': 'str',        # required
                    }
                ],
                'user_id': 'str',                # injected from auth (required)
                'domain_id': 'str',              # injected from auth (required)
            }
        Returns:
           WorkspaceGroupResponse:
        """
        return self.workspace_group_svc.process_add_users(params, role_type="USER")

    @transaction(permission="identity:WorkspaceGroupUser:write", role_types=["USER"])
    @convert_model
    def remove(
        self, params: WorkspaceGroupUserRemoveRequest
    ) -> Union[WorkspaceGroupResponse, dict]:
        """Remove workspace group user
        Args:
            params (WorkspaceGroupUserRemoveRequest): {
                'workspace_group_id': 'str',        # required
                'users': [
                    {
                        'user_id': 'str',           # required
                    }
                ],
                'user_id': 'str',                   # injected from auth (required)
                'domain_id': 'str',                 # injected from auth (required)
            }
        Returns:
            WorkspaceGroupResponse:
        """
        return self.workspace_group_svc.process_remove_users(params, role_type="USER")

    @transaction(permission="identity:WorkspaceGroupUser:write", role_types=["USER"])
    @convert_model
    def update_role(
        self, params: WorkspaceGroupUserUpdateRoleRequest
    ) -> Union[WorkspaceGroupResponse, dict]:
        """Update workspace group user role
        Args:
            params (WorkspaceGroupUserUpdateRoleRequest): {
                'workspace_group_id': 'str',            # required
                'role_id': 'str',                       # required
                'target_user_id': 'str',                # required
                'user_id': 'str',                       # injected from auth (required)
                'domain_id': 'str',                     # injected from auth (required)
            }
        Returns:
            WorkspaceGroupResponse:
        """
        return self.workspace_group_svc.process_update_role(params, role_type="USER")

    @transaction(permission="identity:WorkspaceGroupUser:read", role_types=["USER"])
    @convert_model
    def find(
        self, params: WorkspaceGroupUserFindRequest
    ) -> Union[WorkspaceGroupUsersSummaryResponse, dict]:
        """Find users in the domain except users in its workspace_group
        Args:
            params (WorkspaceGroupUserFindRequest): {
                'workspace_group_id': 'str',      # required
                'keyword': 'str',
                'state': 'State',
                'page': 'dict',
                'user_id': 'str'                  # injected from auth (required)
                'domain_id': 'str',               # injected from auth (required)
            }
        Returns:
            WorkspaceGroupUsersSummaryResponse:
        """
        workspace_group = self.workspace_group_mgr.get_workspace_group(
            params.domain_id, params.workspace_group_id
        )
        workspace_group_users = workspace_group.users or []
        workspace_group_user_ids = list(
            set([user["user_id"] for user in workspace_group_users])
        )

        query = {
            "filter": [
                {"k": "domain_id", "v": params.domain_id, "o": "eq"},
                {"k": "user_id", "v": workspace_group_user_ids, "o": "not_in"},
            ],
            "sort": [{"key": "user_id"}],
            "page": params.page,
            "only": ["user_id", "name", "state"],
        }

        if params.keyword:
            query["filter_or"] = [
                {"k": "user_id", "v": params.keyword, "o": "contain"},
                {"k": "name", "v": params.keyword, "o": "contain"},
            ]

        if params.state:
            query["filter"].append({"k": "state", "v": params.state, "o": "eq"})

        user_mgr = UserManager()
        user_vos, total_count = user_mgr.list_users(query)

        workspace_group_users_info = [user_vo.to_dict() for user_vo in user_vos]

        return WorkspaceGroupUsersSummaryResponse(
            results=workspace_group_users_info, total_count=total_count
        )

    @transaction(permission="identity:WorkspaceGroupUser:read", role_types=["USER"])
    @convert_model
    def get(
        self, params: WorkspaceGroupUserGetRequest
    ) -> Union[WorkspaceGroupResponse, dict]:
        """Get workspace groups
        Args:
            params (WorkspaceGroupUserGetRequest): {
                'workspace_group_id': 'str',     # required
                'user_id': 'str'                 # injected from auth (required)
                'domain_id': 'str',              # injected from auth (required)
            }
        Returns:
            WorkspaceGroupResponse:
        """
        workspace_group_vo, workspace_group_user_ids = (
            self.workspace_group_mgr.get_workspace_group_with_users(
                params.domain_id, params.workspace_group_id
            )
        )

        workspace_group_dict = (
            self.workspace_group_svc.add_user_name_and_state_to_users(
                workspace_group_user_ids, workspace_group_vo, params.domain_id
            )
        )
        return WorkspaceGroupResponse(**workspace_group_dict)

    @transaction(permission="identity:WorkspaceGroupUser.read", role_types=["USER"])
    @append_query_filter(["workspace_group_id", "name", "domain_id"])
    @append_keyword_filter(["workspace_group_id", "name"])
    @convert_model
    def list(
        self, params: WorkspaceGroupUserSearchQueryRequest
    ) -> Union[WorkspaceGroupsResponse, dict]:
        """List workspace group users
        Args:
            params (WorkspaceGroupUserSearchQueryRequest): {
                'query': 'dict (spaceone.api.core.v1.Query)',
                'workspace_group_id': 'str',
                'name': 'str',
                'created_by': 'str',
                'updated_by': 'str',
                'user_id': 'str'                         # injected from auth (required)
                'domain_id': 'str',                      # injected from auth (required)
            }
        Returns:
            WorkspaceGroupsResponse:
        """
        query = params.query

        workspace_group_vos, total_count = (
            self.workspace_group_mgr.list_workspace_groups(query)
        )
        workspace_groups_info = self.workspace_group_svc.get_workspace_groups_info(
            workspace_group_vos, params.domain_id
        )

        return WorkspaceGroupsResponse(
            results=workspace_groups_info, total_count=total_count
        )

    @transaction(permission="identity:WorkspaceGroup.read", role_types=["USER"])
    @append_query_filter(["user_id", "workspace_group_id", "domain_id"])
    @convert_model
    def stat(self, params: WorkspaceGroupUserStatQueryRequest) -> dict:
        """Stat workspace group users
        Args:
            params (WorkspaceGroupUserStatQueryRequest): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)', # required
                'workspace_group_id': 'str',           # required
                'user_id': 'str'                       # injected from auth (required)
                'domain_id': 'str',                    # injected from auth (required)
            }
        Returns:
            dict: {
                'results': 'list',
                'total_count': 'int'
            }
        """
        query = params.query or {}

        return self.workspace_group_user_mgr.stat_workspace_group_users(
            query, params.workspace_group_id, params.domain_id
        )



import logging
from typing import Union

from spaceone.core.service import *
from spaceone.core.service.utils import *
from spaceone.core.error import *

from spaceone.identity.model.provider.request import *
from spaceone.identity.model.provider.response import *
from spaceone.identity.manager.account_collector_plugin_manager import (
    AccountCollectorPluginManager,
)
from spaceone.identity.manager.provider_manager import ProviderManager

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class ProviderService(BaseService):
    resource = "Provider"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.provider_mgr = ProviderManager()
        self.ac_plugin_mgr = AccountCollectorPluginManager()

    @transaction(permission="identity:Provider.write", role_types=["DOMAIN_ADMIN"])
    @convert_model
    def create(self, params: ProviderCreateRequest) -> Union[ProviderResponse, dict]:
        """create provider

        Args:
            params (ProviderCreateRequest): {
                'provider': 'str',      # required
                'name': 'str',          # required
                'alias': 'str',
                'plugin_info: 'dict'
                'color': 'str',
                'icon': 'str',
                'order': 'int',
                'options': 'dict',
                'tags': 'dict',
                'domain_id': 'str'      # injected from auth (required)
            }

        Returns:
            ProviderResponse:
        """

        self._check_sync_option_with_plugin_info(params.options, params.plugin_info)

        provider_vo = self.provider_mgr.create_provider(params.dict())
        return ProviderResponse(**provider_vo.to_dict())

    @transaction(permission="identity:Provider.write", role_types=["DOMAIN_ADMIN"])
    @convert_model
    def update(self, params: ProviderUpdateRequest) -> Union[ProviderResponse, dict]:
        """update provider

        Args:
            params (ProviderUpdateRequest): {
                'provider': 'str',      # required
                'name': 'str',
                'alias': 'str',
                'color': 'str',
                'icon': 'str',
                'order': 'int',
                'options': 'dict',
                'tags': 'dict',
                'domain_id': 'str'      # injected from auth (required)
            }

        Returns:
            ProviderResponse:
        """

        provider_vo = self.provider_mgr.get_provider(params.provider, params.domain_id)
        if provider_vo.is_managed:
            raise ERROR_PERMISSION_DENIED()

        provider_vo = self.provider_mgr.update_provider_by_vo(
            params.dict(exclude_unset=True), provider_vo
        )

        return ProviderResponse(**provider_vo.to_dict())

    @transaction(
        permission="identity:TrustedAccount.write",
        role_types=["DOMAIN_ADMIN"],
    )
    @convert_model
    def update_plugin(
        self, params: ProviderUpdatePluginRequest
    ) -> Union[ProviderResponse, dict]:
        """update provider plugin info
        Args:
            params (ProviderUpdatePluginRequest): {
                'provider': 'str',      # required
                'version': 'str',
                'options': 'dict',
                'upgrade_mode': 'str',
                'domain_id': 'str'      # injected from auth
            }


        """

        provider = params.provider
        domain_id = params.domain_id
        version = params.version
        options = params.options or {}
        upgrade_mode = params.upgrade_mode

        provider = self.provider_mgr.get_provider(provider, domain_id)

        self.check_update_provider_plugin_info(provider.options)

        plugin_info = provider.plugin_info
        if version:
            plugin_info["version"] = version

        if options:
            plugin_info["options"] = options

        if upgrade_mode:
            plugin_info["upgrade_mode"] = upgrade_mode

        (
            endpoint,
            updated_version,
        ) = self.ac_plugin_mgr.get_account_collector_plugin_endpoint(
            plugin_info, domain_id
        )

        if updated_version:
            plugin_info["version"] = updated_version

        options = plugin_info.get("options", {})
        plugin_metadata = self._init_plugin(endpoint, options, domain_id)

        plugin_info["metadata"] = plugin_metadata
        _LOGGER.debug(f"[update_plugin] plugin_info: {plugin_info}")

        provider = self.provider_mgr.update_provider_by_vo(
            {"plugin_info": plugin_info}, provider
        )
        return ProviderResponse(**provider.to_dict())

    @transaction(permission="identity:Provider.write", role_types=["DOMAIN_ADMIN"])
    @convert_model
    def delete(self, params: ProviderDeleteRequest) -> None:
        """delete provider

        Args:
            params (ProviderDeleteRequest): {
                'provider': 'str',      # required
                'domain_id': 'str'      # injected from auth (required)
            }

        Returns:
            None
        """

        provider_vo = self.provider_mgr.get_provider(params.provider, params.domain_id)
        if provider_vo.is_managed:
            raise ERROR_PERMISSION_DENIED()

        self.provider_mgr.delete_provider_by_vo(provider_vo)

    @transaction(
        permission="identity:Provider.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def get(self, params: ProviderGetRequest) -> Union[ProviderResponse, dict]:
        """delete provider

        Args:
            params (ProviderGetRequest): {
                'provider': 'str',      # required
                'domain_id': 'str'      # injected from auth (required)
            }

        Returns:
            ProviderResponse:
        """

        provider_vo = self.provider_mgr.get_provider(params.provider, params.domain_id)
        return ProviderResponse(**provider_vo.to_dict())

    @transaction(
        permission="identity:Provider.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @append_query_filter(["provider", "name", "alias", "is_managed", "domain_id"])
    @append_keyword_filter(["provider", "name"])
    @convert_model
    def list(
        self, params: ProviderSearchQueryRequest
    ) -> Union[ProvidersResponse, dict]:
        """list providers

        Args:
            params (ProviderSearchQueryRequest): {
                'query': 'dict (spaceone.api.core.v1.Query)',
                'provider': 'str',
                'name': 'str',
                'alias': 'str',
                'is_managed': 'bool',
                'domain_id': 'str'      # injected from auth (required)
            }

        Returns:
            ProvidersResponse:
        """

        query = params.query or {}
        provider_vos, total_count = self.provider_mgr.list_providers(
            query, params.domain_id
        )

        providers_info = [provider_vo.to_dict() for provider_vo in provider_vos]
        return ProvidersResponse(results=providers_info, total_count=total_count)

    @transaction(
        permission="identity:Provider.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @append_query_filter(["domain_id"])
    @append_keyword_filter(["provider", "name"])
    @convert_model
    def stat(self, params: ProviderStatQueryRequest) -> dict:
        """stat providers

        Args:
            params (ProviderStatQueryRequest): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)', # required
                'domain_id': 'str'    # injected from auth (required)
            }

        Returns:
            dict: {
                'results': 'list',
                'total_count': 'int'
            }

        """

        query = params.query or {}
        return self.provider_mgr.stat_providers(query)

    def _init_plugin(self, endpoint: str, options: dict, domain_id: str) -> dict:
        self.ac_plugin_mgr.initialize(endpoint)
        return self.ac_plugin_mgr.init_plugin(domain_id, options)

    @staticmethod
    def _check_sync_option_with_plugin_info(
        provider_options: Union[dict, None],
        plugin_info: Union[Plugin, dict, None],
    ) -> None:
        if provider_options:
            if provider_options.get("support_auto_sync"):
                if not provider_options.get("support_trusted_account"):
                    raise ERROR_INVALID_PARAMETER(
                        key="options.support_trusted_account",
                        reason="Provider must support 'Trusted Account' to support 'Service Account' auto sync.",
                    )
                if not plugin_info:
                    raise ERROR_INVALID_PARAMETER(
                        key="plugin_info",
                        reason="Plugin info is required for 'Service Account' auto sync.",
                    )

            if plugin_info:
                if not (
                    provider_options.get("support_trusted_account")
                    or not provider_options.get("support_auto_sync")
                ):
                    raise ERROR_INVALID_PARAMETER(
                        key="options",
                        reason="Both support_trusted_account and support_auto_sync must be enabled.",
                    )

    @staticmethod
    def check_update_provider_plugin_info(provider_options: dict):
        if not provider_options.get("support_trusted_account"):
            raise ERROR_INVALID_PARAMETER(
                key="options",
                reason="Provider support_trusted_account option is disabled.",
            )
        if not provider_options["support_auto_sync"]:
            raise ERROR_INVALID_PARAMETER(
                key="options", reason="Provider support_auto_sync option is disabled."
            )



from spaceone.identity.plugin.external_auth.lib.server import ExternalAuthPluginServer

app = ExternalAuthPluginServer()


@app.route("ExternalAuth.init")
def external_auth_init(params: dict) -> dict:
    """init plugin by options

    Args:
        params (ExternalAuthInitRequest): {
            'options': 'dict',    # Required
            'domain_id': 'str'    # Required
        }

    Returns:
        PluginResponse: {
            'metadata': 'dict'
        }
    """
    pass


@app.route("ExternalAuth.authorize")
def external_auth_authorize(params: dict) -> dict:
    """ExternalAuth authorize

    Args:
        params (ExternalAuthAuthorizeRequest): {
            'options': 'dict',          # Required
            'schema_id': 'str',
            'secret_data': 'dict',      # Required
            'credentials': 'dict',      # Required
            'domain_id': 'str'          # Required
        }

    Returns:
        UserResponse: {
            'state': 'str',
            'user_id': 'str',
            'name': 'str',
            'email': 'str',
            'mobile': 'str',
            'group': 'str',
        }
    """
    pass



import logging
import os

from jinja2 import Environment, FileSystemLoader, select_autoescape

from spaceone.core import config
from spaceone.core.manager import BaseManager
from spaceone.identity.connector.smtp_connector import SMTPConnector

_LOGGER = logging.getLogger(__name__)

TEMPLATE_PATH = os.path.join(os.path.dirname(__file__), f"../template")
JINJA_ENV = Environment(
    loader=FileSystemLoader(searchpath=TEMPLATE_PATH), autoescape=select_autoescape()
)

LANGUAGE_MAPPER = {
    "default": {
        "reset_password": "Reset your password",
        "temp_password": "Your password has been changed",
        "verify_email": "Verify your notification email",
        "invite_external_user": "You've been invited to join.",
    },
    "ko": {
        "reset_password": "비밀번호 재설정 안내",
        "temp_password": "임시 비밀번호 발급 안내",
        "verify_email": "알림전용 이메일 계정 인증 안내",
        "invite_external_user": "계정 초대 안내.",
    },
    "en": {
        "reset_password": "Reset your password",
        "temp_password": "Your password has been changed",
        "verify_email": "Verify your notification email",
        "invite_external_user": "You've been invited to join.",
    },
    "ja": {
        "reset_password": "パスワードリセットのご案内",
        "temp_password": "仮パスワード発行のご案内",
        "verify_email": "通知メールアカウント認証のご案内",
        "invite_external_user": "参加するように招待されました",
    },
}


class EmailManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.smtp_connector = SMTPConnector()

    def send_reset_password_email(self, user_id, email, reset_password_link, language):
        service_name = self._get_service_name()
        language_map_info = LANGUAGE_MAPPER.get(language, "default")

        template = JINJA_ENV.get_template(
            f"reset_pwd_link_when_pw_forgotten_{language}.html"
        )
        email_contents = template.render(
            user_name=user_id,
            reset_password_link=reset_password_link,
            service_name=service_name,
        )
        subject = f'[{service_name}] {language_map_info["reset_password"]}'

        self.smtp_connector.send_email(email, subject, email_contents)

    def send_temporary_password_email(
        self, user_id, email, console_link, temp_password, language
    ):
        service_name = self._get_service_name()
        language_map_info = LANGUAGE_MAPPER.get(language, "default")

        template = JINJA_ENV.get_template(f"temp_pwd_when_pw_forgotten_{language}.html")
        email_contents = template.render(
            user_name=user_id,
            temp_password=temp_password,
            service_name=service_name,
            login_link=console_link,
        )
        subject = f'[{service_name}] {language_map_info["temp_password"]}'

        self.smtp_connector.send_email(email, subject, email_contents)

    def send_reset_password_email_when_user_added(
        self, user_id, email, reset_password_link, language
    ):
        service_name = self._get_service_name()
        language_map_info = LANGUAGE_MAPPER.get(language, "default")

        template = JINJA_ENV.get_template(
            f"reset_pwd_link_when_user_added_{language}.html"
        )
        email_contents = template.render(
            user_name=user_id,
            reset_password_link=reset_password_link,
            service_name=service_name,
        )
        subject = f'[{service_name}] {language_map_info["reset_password"]}'

        self.smtp_connector.send_email(email, subject, email_contents)

    def send_temporary_password_email_when_user_added(
        self, user_id, email, console_link, temp_password, language
    ):
        service_name = self._get_service_name()
        language_map_info = LANGUAGE_MAPPER.get(language, "default")

        template = JINJA_ENV.get_template(f"temp_pwd_when_user_added_{language}.html")
        email_contents = template.render(
            user_name=user_id,
            temp_password=temp_password,
            service_name=service_name,
            login_link=console_link,
        )
        subject = f'[{service_name}] {language_map_info["temp_password"]}'

        self.smtp_connector.send_email(email, subject, email_contents)

    def send_invite_email_when_external_user_added(
        self,
        user_id: str,
        email: str,
        console_link: str,
        language: str,
        external_auth_provider: str = "EXTERNAL",
    ):
        service_name = self._get_service_name()
        language_map_info = LANGUAGE_MAPPER.get(language, "default")

        template = JINJA_ENV.get_template(f"sso_invite_user_link_{language}.html")

        email_contents = template.render(
            user_name=user_id,
            auth_type=external_auth_provider,
            service_name=service_name,
            login_link=console_link,
        )

        subject = f'[{service_name}] {language_map_info["invite_external_user"]}'
        self.smtp_connector.send_email(email, subject, email_contents)

    def send_verification_email(self, user_id, email, verification_code, language):
        service_name = self._get_service_name()
        language_map_info = LANGUAGE_MAPPER.get(language, "default")

        template = JINJA_ENV.get_template(f"verification_code_{language}.html")
        email_contents = template.render(
            user_name=user_id,
            verification_code=verification_code,
            service_name=service_name,
        )
        subject = f'[{service_name}] {language_map_info["verify_email"]}'

        self.smtp_connector.send_email(email, subject, email_contents)

    @staticmethod
    def _get_service_name():
        return config.get_global("EMAIL_SERVICE_NAME")



from typing import Union
from pydantic import BaseModel


__all__ = ["ExternalAuthInitRequest", "ExternalAuthAuthorizeRequest"]


class ExternalAuthInitRequest(BaseModel):
    options: dict
    domain_id: str


class ExternalAuthAuthorizeRequest(BaseModel):
    options: dict
    schema_id: Union[dict, None] = None
    secret_data: dict
    credentials: dict
    domain_id: str



from typing import Dict, List, Literal, Union

from pydantic import BaseModel

__all__ = [
    "WorkspaceGroupCreateRequest",
    "WorkspaceGroupUpdateRequest",
    "WorkspaceGroupDeleteRequest",
    "WorkspaceGroupAddUsersRequest",
    "WorkspaceGroupRemoveUsersRequest",
    "WorkspaceGroupUpdateRoleRequest",
    "WorkspaceGroupGetRequest",
    "WorkspaceGroupSearchQueryRequest",
    "WorkspaceGroupStatQueryRequest",
    "State",
]

State = Literal["ENABLED", "DISABLED", "PENDING"]


class WorkspaceGroupCreateRequest(BaseModel):
    name: str
    tags: Union[dict, None] = None
    workspace_id: Union[str, None] = None
    domain_id: str


class WorkspaceGroupUpdateRequest(BaseModel):
    workspace_group_id: str
    name: Union[str, None] = None
    tags: Union[dict, None] = None
    workspace_id: Union[str, None] = None
    domain_id: str


class WorkspaceGroupDeleteRequest(BaseModel):
    workspace_group_id: str
    workspace_id: Union[str, None] = None
    domain_id: str


class WorkspaceGroupAddUsersRequest(BaseModel):
    workspace_group_id: str
    users: List[Dict[str, str]]
    workspace_id: Union[str, None] = None
    domain_id: str


class WorkspaceGroupRemoveUsersRequest(BaseModel):
    workspace_group_id: str
    users: List[Dict[str, str]]
    workspace_id: Union[str, None] = None
    domain_id: str


class WorkspaceGroupUpdateRoleRequest(BaseModel):
    workspace_group_id: str
    user_id: str
    role_id: str
    workspace_id: Union[str, None] = None
    domain_id: str


class WorkspaceGroupGetRequest(BaseModel):
    workspace_group_id: str
    workspace_id: Union[str, None] = None
    domain_id: str


class WorkspaceGroupSearchQueryRequest(BaseModel):
    query: Union[dict, None] = None
    workspace_group_id: Union[str, None] = None
    name: Union[str, None] = None
    created_by: Union[str, None] = None
    updated_by: Union[str, None] = None
    workspace_id: Union[str, None] = None
    domain_id: str


class WorkspaceGroupStatQueryRequest(BaseModel):
    query: dict
    workspace_id: Union[str, None] = None
    domain_id: str



import logging
import os

from jinja2 import Environment, FileSystemLoader, select_autoescape

from spaceone.core import config
from spaceone.identity.connector.smtp_connector import SMTPConnector
from spaceone.identity.manager.mfa_manager.base import MFAManager

_LOGGER = logging.getLogger(__name__)

TEMPLATE_PATH = os.path.join(os.path.dirname(__file__), f"../../template")
JINJA_ENV = Environment(
    loader=FileSystemLoader(searchpath=TEMPLATE_PATH), autoescape=select_autoescape()
)

LANGUAGE_MAPPER = {
    "default": {
        "verify_mfa_email": "Verify your MFA email",
        "authentication_mfa_email": "Your multi-factor authentication code.",
    },
    "ko": {
        "verify_mfa_email": "MFA 이메일 계정 인증 안내",
        "authentication_mfa_email": "MFA 인증 코드 발송",
    },
    "en": {
        "verify_mfa_email": "Verify your MFA email",
        "authentication_mfa_email": "Your multi-factor authentication code.",
    },
    "ja": {
        "verify_mfa_email": "MFAメールアカウント認証のご案内",
        "authentication_mfa_email": "MFA認証コード送信",
    },
}


class EmailMFAManager(MFAManager):
    mfa_type = "EMAIL"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.smtp_connector = SMTPConnector()

    def enable_mfa(self, user_id, domain_id, user_mfa, language):
        self.send_mfa_verify_email(
            user_id, domain_id, user_mfa["options"].get("email"), language
        )

    def disable_mfa(self, user_id, domain_id, user_mfa, language):
        self.send_mfa_verify_email(
            user_id, domain_id, user_mfa["options"].get("email"), language
        )

    def confirm_mfa(self, credentials: dict, verify_code: str):
        return self.check_mfa_verify_code(credentials, verify_code)

    def send_mfa_verify_email(self, user_id, domain_id, email, language):
        service_name = self._get_service_name()
        language_map_info = LANGUAGE_MAPPER.get(language, "default")
        credentials = {"user_id": user_id, "domain_id": domain_id}
        verify_code = self.create_mfa_verify_code(user_id, domain_id, credentials)

        template = JINJA_ENV.get_template(f"verification_MFA_code_{language}.html")
        email_contents = template.render(
            user_name=user_id, verification_code=verify_code, service_name=service_name
        )
        subject = f'[{service_name}] {language_map_info["verify_mfa_email"]}'

        self.smtp_connector.send_email(email, subject, email_contents)

    def send_mfa_authentication_email(
        self, user_id: str, domain_id: str, email: str, language: str, credentials: dict
    ):
        service_name = self._get_service_name()
        language_map_info = LANGUAGE_MAPPER.get(language, "default")
        verify_code = self.create_mfa_verify_code(user_id, domain_id, credentials)

        template = JINJA_ENV.get_template(f"authentication_code_{language}.html")
        email_contents = template.render(
            user_name=user_id,
            authentication_code=verify_code,
            service_name=service_name,
        )
        subject = f'[{service_name}] {language_map_info["authentication_mfa_email"]}'

        self.smtp_connector.send_email(email, subject, email_contents)

    @staticmethod
    def _get_service_name():
        return config.get_global("EMAIL_SERVICE_NAME")



from spaceone.core.pygrpc.server import GRPCServer
from spaceone.identity.plugin.external_auth.interface.grpc.external_auth import (
    ExternalAuth,
)

_all_ = ["app"]

app = GRPCServer()
app.add_service(ExternalAuth)



from typing import Union, Literal
from pydantic import BaseModel


__all__ = [
    "AgentCreateRequest",
    "AgentEnableRequest",
    "AgentDisableRequest",
    "AgentRegenerateRequest",
    "AgentDeleteRequest",
    "AgentGetRequest",
    "AgentSearchQueryRequest",
]

State = Literal["ENABLED", "DISABLED", "EXPIRED"]
RoleType = Literal["DOMAIN_ADMIN", "WORKSPACE_OWNER"]


class AgentCreateRequest(BaseModel):
    service_account_id: str
    options: Union[dict, None] = None
    workspace_id: str
    domain_id: str
    user_projects: Union[list, None] = None


class AgentEnableRequest(BaseModel):
    service_account_id: str
    workspace_id: str
    domain_id: str
    user_projects: Union[list, None] = None


class AgentDisableRequest(BaseModel):
    service_account_id: str
    workspace_id: str
    domain_id: str
    user_projects: Union[list, None] = None


class AgentRegenerateRequest(BaseModel):
    service_account_id: str
    workspace_id: str
    domain_id: str
    user_projects: Union[list, None] = None


class AgentDeleteRequest(BaseModel):
    service_account_id: str
    workspace_id: str
    domain_id: str
    user_projects: Union[list, None] = None


class AgentGetRequest(BaseModel):
    service_account_id: str
    workspace_id: str
    domain_id: str
    user_projects: Union[list, None] = None


class AgentSearchQueryRequest(BaseModel):
    query: Union[dict, None] = None
    agent_id: Union[str, None] = None
    state: Union[State, None] = None
    service_account_id: Union[str, None] = None
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None



import logging
import random
from datetime import datetime, timedelta
from typing import Union, List

from spaceone.core.service import *
from spaceone.core.service.utils import *
from spaceone.core import config

from spaceone.identity.conf.global_conf import WORKSPACE_COLORS_NAME
from spaceone.identity.error.error_job import *
from spaceone.identity.manager.account_collector_plugin_manager import (
    AccountCollectorPluginManager,
)
from spaceone.identity.manager.job_manager import JobManager
from spaceone.identity.manager.project_manager import ProjectManager
from spaceone.identity.manager.project_group_manager import ProjectGroupManager
from spaceone.identity.manager.provider_manager import ProviderManager
from spaceone.identity.manager.schema_manager import SchemaManager
from spaceone.identity.manager.service_account_manager import ServiceAccountManager
from spaceone.identity.manager.secret_manager import SecretManager
from spaceone.identity.manager.trusted_account_manager import TrustedAccountManager
from spaceone.identity.manager.workspace_manager import WorkspaceManager
from spaceone.identity.manager.domain_manager import DomainManager
from spaceone.identity.manager.config_manager import ConfigManager
from spaceone.identity.manager.cost_analysis_manager import CostAnalysisManager
from spaceone.identity.model.project.database import Project
from spaceone.identity.model.project_group.database import ProjectGroup
from spaceone.identity.model.provider.database import Provider
from spaceone.identity.model.service_account.database import ServiceAccount
from spaceone.identity.model.trusted_account.database import TrustedAccount
from spaceone.identity.model.job.database import Job
from spaceone.identity.model.job.request import *
from spaceone.identity.model.job.response import *
from spaceone.identity.model.workspace.database import Workspace

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class JobService(BaseService):
    resource = "Job"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.job_mgr = JobManager()
        self.trusted_account_mgr = TrustedAccountManager()
        self.provider_mgr = ProviderManager()
        self.account_collector_plugin_mgr = AccountCollectorPluginManager()
        self.workspace_mgr = WorkspaceManager()
        self.service_account_mgr = ServiceAccountManager()
        self.project_mgr = ProjectManager()
        self.project_group_mgr = ProjectGroupManager()

    @transaction(exclude=["authentication", "authorization", "mutation"])
    def create_jobs_by_trusted_account(self, params: dict):
        """Create jobs by trusted account
        Args:
            params (dict): {
                'params':
            }
        Returns:
            None:
        """

        current_hour = params.get("current_hour", datetime.utcnow().hour)

        for trusted_account_vo in self._get_all_schedule_enabled_trusted_accounts(
            current_hour
        ):
            try:
                self.create_service_account_job(trusted_account_vo, {})
            except Exception as e:
                _LOGGER.error(
                    f"[create_jobs_by_trusted_account] sync error: {e}", exc_info=True
                )

    @transaction(
        permission="identity:Job.write", role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"]
    )
    @convert_model
    def delete(self, params: JobDeleteRequest) -> None:
        """Delete job
        Args:
            params (JobDeleteRequest): {
                'job_id': 'str',            # required
                'workspace_id': 'str',      # injected from auth (required)
                'domain_id': 'str'          # injected from auth (required)
            }
        Returns:
            None:
        """

        job_vo = self.job_mgr.get_job(
            params.domain_id,
            params.job_id,
            params.workspace_id,
        )

        self.job_mgr.delete_job_by_vo(job_vo)

    @transaction(
        permission="identity:Job.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @convert_model
    def get(self, params: JobGetRequest) -> JobResponse:
        """Get job
        Args:
            params (JobGetRequest): {
                'job_id': 'str',            # required
                'workspace_id': 'str',      # injected from auth (required)
                'domain_id': 'str'          # injected from auth (required)
            }
        Returns:
            JobInfo:
        """

        job_id = params.job_id
        workspace_id = params.workspace_id
        domain_id = params.domain_id

        job_vo = self.job_mgr.get_job(domain_id, job_id, workspace_id)

        return JobResponse(**job_vo.to_dict())

    @transaction(
        permission="identity:Job.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @append_query_filter(
        [
            "job_id",
            "status",
            "trusted_account_id",
            "plugin_id",
            "workspace_id",
            "domain_id",
        ]
    )
    @append_keyword_filter(["job_id", "status"])
    @convert_model
    def list(self, params: JobSearchQueryRequest) -> Union[JobsResponse, dict]:
        """List jobs
        Args:
            params (JobSearchQueryRequest): {
                'query': 'dict',
                'job_id': 'str',
                'status': 'str',
                'trusted_account_id': 'str',
                'plugin_id': 'str',
                'workspace_id': 'str',
                'domain_id': 'str'          # injected from auth (required)
            }
        Returns:
            JobsResponse:
        """

        query = params.query or {}

        job_vos, total_count = self.job_mgr.list_jobs(query)
        jobs_info = [job_vo.to_dict() for job_vo in job_vos]

        return JobsResponse(results=jobs_info, total_count=total_count)

    @transaction(
        permission="identity:Job.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @append_query_filter(["workspace_id", "domain_id"])
    @append_keyword_filter(["job_id"])
    @convert_model
    def stat(self, params: JobStatQueryRequest) -> dict:
        """Stat jobs
        Args:
            params (JobStatQueryRequest): {
                'domain_id': 'str' # injected from auth (required)
            }
        Returns:
            dict:
        """

        query = params.query or {}

        return self.job_mgr.stat_jobs(query)

    @transaction(exclude=["authentication", "authorization", "mutation"])
    def check_dormancy(self, params: dict) -> None:
        """Check dormancy by domains
        Args:
            params (dict): {}
        Returns:
            None:
        """

        domain_mgr = DomainManager()

        domain_vos = domain_mgr.filter_domains(state="ENABLED")
        for domain_vo in domain_vos:
            self.job_mgr.push_dormancy_job({"domain_id": domain_vo.domain_id})

    @transaction(exclude=["authentication", "authorization", "mutation"])
    def check_dormancy_by_domain(self, params: dict) -> None:
        """Check dormancy by domains
        Args:
            params (dict): {
                'domain_id': 'str'
            }
        Returns:
            None:
        """

        domain_id = params["domain_id"]
        dormancy_settings = self._get_dormancy_settings(domain_id)
        dormancy_state = dormancy_settings.get("state")
        threshold = dormancy_settings.get("cost")
        dormancy_send_email = dormancy_settings.get("send_email")

        cost_analysis_mgr = CostAnalysisManager()

        workspace_vos = self.workspace_mgr.filter_workspaces(domain_id=domain_id)
        for workspace_vo in workspace_vos:
            if self._is_dormancy_updated(workspace_vo):
                continue

            cost_info = workspace_vo.cost_info or {}
            before_month_cost = cost_info.get("month", 0)
            before_dormant_ttl = workspace_vo.dormant_ttl or -1
            before_is_dormant = workspace_vo.is_dormant

            update_params = {
                "is_dormant": workspace_vo.is_dormant,
                "dormant_ttl": workspace_vo.dormant_ttl,
                "service_account_count": workspace_vo.service_account_count,
            }

            workspace_id = workspace_vo.workspace_id
            service_account_vos = self.service_account_mgr.filter_service_accounts(
                domain_id=domain_id, workspace_id=workspace_id
            )

            update_params["service_account_count"] = service_account_vos.count()

            month_cost = self._get_this_month_cost(
                cost_analysis_mgr, domain_id, workspace_id
            )

            day_cost = month_cost - before_month_cost
            if day_cost < 0:
                day_cost = 0

            update_params["cost_info"] = {"month": month_cost, "day": day_cost}

            if dormancy_state == "ENABLED" and before_dormant_ttl >= 0:
                if day_cost <= threshold:
                    _LOGGER.debug(
                        f"[check_dormancy_by_domain] change dormant: {workspace_vo.name}"
                    )
                    update_params["is_dormant"] = True

                    if before_dormant_ttl > 0:
                        dormant_ttl = before_dormant_ttl - 1
                        if dormancy_send_email:
                            # Send Email
                            if dormant_ttl == 0:
                                pass
                            else:
                                pass

                        update_params["dormant_ttl"] = dormant_ttl
                else:
                    if before_is_dormant:
                        _LOGGER.debug(
                            f"[check_dormancy_by_domain] change active: {workspace_vo.name}"
                        )

                    update_params["is_dormant"] = False
                    update_params["dormant_ttl"] = 3

            update_params["dormant_updated_at"] = datetime.utcnow()
            self.workspace_mgr.update_workspace_by_vo(update_params, workspace_vo)

    @staticmethod
    def _is_dormancy_updated(workspace_vo: Workspace) -> bool:
        if not workspace_vo.dormant_updated_at:
            return False

        today = datetime.utcnow().strftime("%Y-%m-%d")
        last_dormant_updated_date = workspace_vo.dormant_updated_at.strftime("%Y-%m-%d")

        if today == last_dormant_updated_date:
            _LOGGER.debug(
                f"[_is_dormancy_updated] dormant has already been updated: {workspace_vo.name}"
            )
            return True
        else:
            return False

    @staticmethod
    def _get_this_month_cost(
        cost_analysis_mgr: CostAnalysisManager,
        domain_id: str,
        workspace_id: str,
    ) -> float:
        system_token = config.get_global("TOKEN")
        now = datetime.utcnow()
        report_month = now.strftime("%Y-%m")

        # Get Monthly Cost
        params = {
            "status": "IN_PROGRESS",
            "query": {
                "filter": [
                    {"k": "report_month", "v": report_month, "o": "eq"},
                    {"k": "workspace_id", "v": workspace_id, "o": "eq"},
                ]
            },
        }

        response = cost_analysis_mgr.list_cost_reports(
            params, token=system_token, x_domain_id=domain_id
        )
        results = response.get("results", [])
        if len(results) > 0:
            currency = results[0]["currency"]
            return results[0]["cost"][currency]
        else:
            return 0

    @staticmethod
    def _get_dormancy_settings(domain_id: str) -> dict:
        settings_key = config.get_global(
            "DORMANCY_SETTINGS_KEY", "identity:dormancy:workspace"
        )

        config_mgr = ConfigManager()
        system_token = config.get_global("TOKEN")
        response = config_mgr.list_domain_configs(
            params={"name": settings_key}, token=system_token, x_domain_id=domain_id
        )
        if response.get("total_count", 0) > 0:
            dormancy_settings = response["results"][0]["data"]
            dormancy_state = "ENABLED"
            dormancy_send_email = dormancy_settings.get("send_email", False)
            dormancy_cost = dormancy_settings.get("cost", 0)
        else:
            dormancy_state = "DISABLED"
            dormancy_send_email = False
            dormancy_cost = 0

        return {
            "state": dormancy_state,
            "send_email": dormancy_send_email,
            "cost": dormancy_cost,
        }

    @transaction(exclude=["authentication", "authorization", "mutation"])
    def sync_service_accounts(self, params: dict) -> None:
        """Sync account data
        Args:
            params (dict): {
                    'job_id': 'str',
                    'trusted_account_id': 'str',
                    'trusted_secret_id': 'str',
                    'secret_data': 'dict',
                    'workspace_id': 'str',
                    'domain_id': 'str'
                    'options': 'dict'
            }
        Returns:
            None:
        """

        trusted_secret_id = params["trusted_secret_id"]
        secret_data = params["secret_data"]
        trusted_account_id = params["trusted_account_id"]
        job_id = params["job_id"]
        workspace_id = params.get("workspace_id")
        domain_id = params["domain_id"]

        schema_mgr = SchemaManager()

        trusted_account_vo: TrustedAccount = (
            self.trusted_account_mgr.get_trusted_account(
                trusted_account_id, domain_id, workspace_id
            )
        )
        job_vo: Job = self.job_mgr.get_job(domain_id, job_id, workspace_id)
        schema_mgr.get_schema(trusted_account_vo.secret_schema_id, domain_id)

        provider_vo: Provider = self.provider_mgr.get_provider(
            trusted_account_vo.provider, domain_id
        )
        plugin_info = provider_vo.plugin_info

        provider = provider_vo.provider
        sync_options = trusted_account_vo.sync_options or {}
        plugin_options = trusted_account_vo.plugin_options or {}

        if self._is_job_failed(job_id, domain_id, job_vo.workspace_id):
            self.job_mgr.change_canceled_status(job_vo)
        else:
            self.job_mgr.change_in_progress_status(job_vo)

            try:
                # Merge plugin options and trusted_account plugin options
                options = plugin_info.get("options", {})
                options.update(plugin_options)
                schema_id = plugin_info.get("schema_id")

                (
                    endpoint,
                    updated_version,
                ) = self.account_collector_plugin_mgr.get_account_collector_plugin_endpoint(
                    plugin_info, domain_id
                )

                self.account_collector_plugin_mgr.initialize(endpoint)
                start_dt = datetime.utcnow()

                is_canceled = False

                response = self.account_collector_plugin_mgr.sync(
                    endpoint, options, secret_data, domain_id, schema_id
                )

                for result in response.get("results", []):
                    location: List[dict] = self._get_location(
                        result, trusted_account_vo.resource_group, sync_options
                    )

                    if trusted_account_vo.resource_group == "DOMAIN":
                        if sync_options.get("single_workspace_id"):
                            workspace_vo = self.workspace_mgr.get_workspace(
                                sync_options.get("single_workspace_id"), domain_id
                            )
                        elif location:
                            location_info = location.pop(0)
                            workspace_vo = self._create_workspace(
                                domain_id, trusted_account_id, location_info
                            )
                        else:
                            _LOGGER.debug(
                                f"[sync_service_accounts] location is empty => SKIP"
                            )
                            continue

                        sync_workspace_id = workspace_vo.workspace_id

                    else:
                        sync_workspace_id = workspace_id

                    parent_group_id = None
                    for location_info in location:
                        project_group_vo = self._create_project_group(
                            domain_id,
                            sync_workspace_id,
                            trusted_account_id,
                            location_info,
                            parent_group_id,
                        )
                        parent_group_id = project_group_vo.project_group_id

                    project_vo = self._create_project(
                        result,
                        domain_id,
                        sync_workspace_id,
                        trusted_account_id,
                        project_group_id=parent_group_id,
                        sync_options=sync_options,
                    )
                    self._create_service_account(
                        result,
                        project_vo,
                        trusted_account_id,
                        trusted_secret_id,
                        provider,
                        sync_options,
                    )

                if self._is_job_failed(job_id, domain_id, job_vo.workspace_id):
                    self.job_mgr.change_canceled_status(job_vo)
                    is_canceled = True

                if not is_canceled:
                    end_dt = datetime.utcnow()
                    _LOGGER.debug(
                        f"[sync_service_accounts] end job ({job_vo.job_id}): {end_dt}"
                    )
                    _LOGGER.debug(
                        f"[sync_service_accounts] total job time ({job_vo.job_id}): {end_dt - start_dt}"
                    )
                    self.job_mgr.change_success_status(job_vo)

            except Exception as e:
                self.job_mgr.change_error_status(job_vo, e)
                _LOGGER.error(f"[sync_service_accounts] sync error: {e}", exc_info=True)

        self._close_job(
            job_id,
            domain_id,
            job_vo.workspace_id,
        )

    def create_service_account_job(
        self, trusted_account_vo: TrustedAccount, job_options: dict
    ) -> Union[Job, dict]:
        resource_group = trusted_account_vo.resource_group
        provider = trusted_account_vo.provider
        trusted_account_id = trusted_account_vo.trusted_account_id

        if resource_group == "DOMAIN":
            workspace_id = "*"
        else:
            workspace_id = trusted_account_vo.workspace_id

        domain_id = trusted_account_vo.domain_id

        provider_vo = self.provider_mgr.get_provider(provider, domain_id)
        plugin_id = provider_vo.plugin_info["plugin_id"]
        ac_plugin_mgr = AccountCollectorPluginManager()
        endpoint = ac_plugin_mgr.get_account_collector_plugin_endpoint_by_vo(
            provider_vo
        )

        options = provider_vo.plugin_info.get("options")
        schema_id = trusted_account_vo.secret_schema_id

        ac_plugin_mgr.initialize(endpoint)

        try:
            trusted_secret_data = self._get_trusted_secret_data(
                trusted_account_vo.trusted_secret_id, domain_id
            )

            if trusted_secret_data:
                schema_mgr = SchemaManager()
                # Check secret_data by schema
                schema_mgr.validate_secret_data_by_schema_id(
                    schema_id, domain_id, trusted_secret_data, "SECRET"
                )
        except Exception as e:
            trusted_secret_data = {}
            _LOGGER.error(
                f"[created_trusted_account_job] get trusted secret error: {e}",
                exc_info=True,
            )

        # Add Job Options
        job_vo = self.job_mgr.create_job(
            resource_group,
            domain_id,
            workspace_id,
            trusted_account_id,
            plugin_id,
            job_options,
        )

        if self._check_duplicate_job(domain_id, trusted_account_id, job_vo):
            self.job_mgr.change_error_status(
                job_vo, ERROR_DUPLICATE_JOB(trusted_account_id=trusted_account_id)
            )
        else:
            try:
                self.job_mgr.push_job(
                    {
                        "job_id": job_vo.job_id,
                        "trusted_account_id": trusted_account_id,
                        "trusted_secret_id": trusted_account_vo.trusted_secret_id,
                        "secret_data": trusted_secret_data,
                        "workspace_id": trusted_account_vo.workspace_id,
                        "domain_id": domain_id,
                        "options": job_options,
                    }
                )
            except Exception as e:
                self.job_mgr.change_error_status(job_vo, e)

        return job_vo

    def _get_all_schedule_enabled_trusted_accounts(self, current_hour: int) -> list:
        query = {
            "filter": [
                {"k": "schedule.state", "v": "ENABLED", "o": "eq"},
                {"k": "schedule.hours", "v": [current_hour], "o": "in"},
            ]
        }
        (
            trusted_account_vos,
            total_count,
        ) = self.trusted_account_mgr.list_trusted_accounts(query)
        _LOGGER.debug(
            f"[_get_all_schedule_enabled_trusted_accounts] scheduled trusted accounts count (UTC {current_hour}: {total_count}"
        )
        return trusted_account_vos

    def _get_trusted_secret_data(self, trusted_secret_id: str, domain_id: str) -> dict:
        secret_mgr: SecretManager = self.locator.get_manager("SecretManager")
        if trusted_secret_id:
            secret_data = secret_mgr.get_trusted_secret_data(
                trusted_secret_id, domain_id
            )
        else:
            secret_data = {}

        return secret_data

    def _check_duplicate_job(
        self,
        domain_id: str,
        trusted_account_id: str,
        this_job_vo: Job,
    ) -> bool:
        query = {
            "filter": [
                {"k": "trusted_account_id", "v": trusted_account_id, "o": "eq"},
                {"k": "workspace_id", "v": this_job_vo.workspace_id, "o": "eq"},
                {"k": "domain_id", "v": domain_id, "o": "eq"},
                {"k": "status", "v": "IN_PROGRESS", "o": "eq"},
                {"k": "job_id", "v": this_job_vo.job_id, "o": "not"},
            ]
        }

        job_vos, total_count = self.job_mgr.list_jobs(query)

        duplicate_job_time = datetime.utcnow() - timedelta(minutes=10)

        for job_vo in job_vos:
            if job_vo.created_at > duplicate_job_time:
                return True
            else:
                self.job_mgr.change_canceled_by_vo(job_vo)
        return False

    def _is_job_failed(
        self, job_id: str, domain_id: str, workspace_id: str = None
    ) -> bool:
        job_vo: Job = self.job_mgr.get_job(domain_id, job_id, workspace_id)

        if job_vo.status in ["CANCELED", "FAILURE"]:
            return True
        else:
            return False

    def _close_job(
        self,
        job_id: str,
        domain_id: str,
        workspace_id: str = None,
    ):
        job_vo: Job = self.job_mgr.get_job(domain_id, job_id, workspace_id)
        if job_vo.status == "IN_PROGRESS":
            self.job_mgr.change_success_status(job_vo)
        elif job_vo.status == "FAILURE":
            self.job_mgr.update_job_by_vo({"finished_at": datetime.utcnow()}, job_vo)

    def _create_workspace(
        self, domain_id: str, trusted_account_id: str, location_info: dict
    ) -> Workspace:
        name = location_info.get("name")
        reference_id = location_info.get("resource_id")

        workspace_vos = self.workspace_mgr.filter_workspaces(
            domain_id=domain_id, name=name
        )

        _LOGGER.debug(
            f"[_create_workspace] 'name': '{name}', 'domain_id': '{domain_id}', count: {len(workspace_vos)}"
        )

        params = {"trusted_account_id": trusted_account_id, "is_managed": True}
        if workspace_vos:
            workspace_vo = workspace_vos[0]

            if not workspace_vo.references:
                params["references"] = [reference_id]
            elif reference_id not in workspace_vo.references:
                params["references"] = workspace_vo.references + [reference_id]

            params["last_synced_at"] = datetime.utcnow()
            _LOGGER.debug(f"[_create_workspace] update workspace: {params}")

            workspace_vo = self.workspace_mgr.update_workspace_by_vo(
                params, workspace_vo
            )

            self._remove_old_reference_id_from_workspace(
                domain_id, workspace_vo.workspace_id, reference_id
            )
        else:
            params.update(
                {
                    "name": name,
                    "tags": self._set_workspace_theme(),
                    "domain_id": domain_id,
                    "last_synced_at": datetime.utcnow(),
                    "references": [reference_id],
                }
            )
            workspace_vo = self.workspace_mgr.create_workspace(params)
        return workspace_vo

    def _create_project_group(
        self,
        domain_id: str,
        workspace_id: str,
        trusted_account_id: str,
        location_info: dict,
        parent_group_id: str = None,
    ) -> ProjectGroup:
        name = location_info["name"]
        reference_id = location_info["resource_id"]

        filter_params = {
            "is_managed": True,
            "reference_id": reference_id,
            "domain_id": domain_id,
            "workspace_id": workspace_id,
        }

        project_group_vos = self.project_group_mgr.filter_project_groups(
            **filter_params
        )

        params = {
            "trusted_account_id": trusted_account_id,
        }
        if parent_group_id:
            params.update({"parent_group_id": parent_group_id})

        if project_group_vos:
            project_group_vo = project_group_vos[0]
            if project_group_vo.name != name:
                params.update({"name": name})

            params.update({"last_synced_at": datetime.utcnow()})
            project_group_vo = self.project_group_mgr.update_project_group_by_vo(
                params, project_group_vo
            )

        else:
            params.update(
                {
                    "name": name,
                    "reference_id": reference_id,
                    "is_managed": True,
                    "domain_id": domain_id,
                    "workspace_id": workspace_id,
                    "last_synced_at": datetime.utcnow(),
                }
            )
            if parent_group_id:
                params["parent_group_id"] = parent_group_id
            project_group_vo = self.project_group_mgr.create_project_group(params)

        return project_group_vo

    def _create_project(
        self,
        result: dict,
        domain_id: str,
        workspace_id: str,
        trusted_account_id: str,
        project_group_id: str = None,
        sync_options: dict = None,
        project_type: str = "PRIVATE",
    ) -> Project:
        name = result["name"]
        reference_id = result["resource_id"]

        params = {
            "domain_id": domain_id,
            "workspace_id": workspace_id,
            "project_type": project_type,
            "reference_id": reference_id,
            "is_managed": True,
        }

        project_vos = self.project_mgr.filter_projects(**params)
        _LOGGER.debug(
            f"[_create_service_account] project_vos: 'name': '{name}', 'params': '{params}' count: {len(project_vos)}"
        )

        if project_group_id:
            params["project_group_id"] = project_group_id

        if project_vos:
            project_vo = project_vos[0]
            if project_vo.name != name:
                params.update({"name": name})

            params.update(
                {
                    "trusted_account_id": trusted_account_id,
                    "last_synced_at": datetime.utcnow(),
                }
            )
            project_vo = self.project_mgr.update_project_by_vo(params, project_vo)
        else:
            params.update({"name": name, "last_synced_at": datetime.utcnow()})
            project_vo = self.project_mgr.create_project(params)
        return project_vo

    def _create_service_account(
        self,
        result: dict,
        project_vo: Project,
        trusted_account_id: str,
        trusted_secret_id: str,
        provider: str,
        sync_options: dict = None,
    ) -> Union[ServiceAccount, None]:
        domain_id = project_vo.domain_id
        workspace_id = project_vo.workspace_id
        project_id = project_vo.project_id
        name = result["name"]
        reference_id = result["resource_id"]
        secret_data = result.get("secret_data", {})
        data = result.get("data", {})
        secret_schema_id = result.get("secret_schema_id")
        tags = result.get("tags", {})

        params = {
            "provider": provider,
            "reference_id": reference_id,
            "is_managed": True,
            "domain_id": domain_id,
            "workspace_id": workspace_id,
            "project_id": project_id,
        }

        service_account_vos = self.service_account_mgr.filter_service_accounts(**params)
        _LOGGER.debug(
            f"[_create_service_account] service_account_vos: {name} {params} count: {len(service_account_vos)}"
        )

        if service_account_vos:
            service_account_vo = service_account_vos[0]
            update_params = {}
            if service_account_vo.name != result["name"]:
                update_params.update({"name": name})

            update_params = {
                "trusted_account_id": trusted_account_id,
                "last_synced_at": datetime.utcnow(),
            }

            service_account_vo = self.service_account_mgr.update_service_account_by_vo(
                update_params, service_account_vo
            )
        else:
            params.update(
                {
                    "name": name,
                    "data": data,
                    "trusted_account_id": trusted_account_id,
                    "tags": tags,
                    "last_synced_at": datetime.utcnow(),
                }
            )
            if secret_schema_id:
                params["schema_id"] = secret_schema_id

            service_account_vo = self.service_account_mgr.create_service_account(params)

        if secret_data:
            secret_mgr: SecretManager = self.locator.get_manager("SecretManager")
            secret_id = service_account_vo.secret_id

            response = secret_mgr.list_secrets({"secret_id": secret_id}, domain_id)
            secret_total_count = response.get("total_count", 0)

            if secret_total_count > 0:
                update_secret_params = {
                    "secret_id": service_account_vo.secret_id,
                    "data": secret_data,
                    "schema_id": secret_schema_id,
                }
                secret_mgr.update_secret_data(
                    update_secret_params, domain_id, workspace_id
                )
            else:

                # Check secret_data by schema
                schema_mgr = SchemaManager()
                schema_mgr.validate_secret_data_by_schema_id(
                    secret_schema_id,
                    service_account_vo.domain_id,
                    secret_data,
                    "TRUSTING_SECRET",
                )

                create_secret_params = {
                    "name": f"{service_account_vo.service_account_id}-secret",
                    "data": secret_data,
                    "resource_group": "PROJECT",
                    "workspace_id": workspace_id,
                    "project_id": project_id,
                    "service_account_id": service_account_vo.service_account_id,
                    "trusted_secret_id": trusted_secret_id,
                    "schema_id": secret_schema_id,
                }
                secret_info = secret_mgr.create_secret(create_secret_params, domain_id)
                service_account_vo = (
                    self.service_account_mgr.update_service_account_by_vo(
                        {"secret_id": secret_info["secret_id"]}, service_account_vo
                    )
                )
        return service_account_vo

    def _remove_old_reference_id_from_workspace(
        self, domain_id: str, workspace_id: str, reference_id: str
    ) -> None:
        query = {
            "filter": [
                {"k": "domain_id", "v": domain_id, "o": "eq"},
                {"k": "workspace_id", "v": workspace_id, "o": "not"},
                {"k": "references", "v": [reference_id], "o": "in"},
            ],
        }
        workspace_vos, _ = self.workspace_mgr.list_workspaces(query)
        for workspace_vo in workspace_vos:
            references = workspace_vo.references
            references.remove(reference_id)
            self.workspace_mgr.update_workspace_by_vo(
                {"references": references}, workspace_vo
            )

    @staticmethod
    def _get_location(result: dict, resource_group: str, sync_options: dict) -> list:
        location = result.get("location", [])
        skip_project_group_option = sync_options.get("skip_project_group")

        if skip_project_group_option:
            if resource_group == "DOMAIN":
                if location:
                    location = [location[0]]
            else:
                location = []

        else:
            if resource_group == "DOMAIN" and not location:
                _LOGGER.debug(
                    f"[_get_location] location is empty: {result} {sync_options} => SKIP"
                )
                # raise ERROR_REQUIRED_PARAMETER(
                #     key="location", reason="location is required"
                # )

        return location

    @staticmethod
    def _set_workspace_theme(tags: dict = None) -> dict:
        theme = random.choice(WORKSPACE_COLORS_NAME)
        if tags:
            tags.update({"theme": theme})
        else:
            tags = {"theme": theme}

        return tags



import logging
import re
from typing import Tuple

import pytz
from mongoengine import QuerySet
from spaceone.core.manager import BaseManager

from spaceone.identity.error.error_user import *
from spaceone.identity.lib.cipher import PasswordCipher
from spaceone.identity.manager.project_manager import ProjectManager
from spaceone.identity.manager.role_binding_manager import RoleBindingManager
from spaceone.identity.manager.user_group_manager import UserGroupManager
from spaceone.identity.manager.workspace_group_manager import \
    WorkspaceGroupManager
from spaceone.identity.model.user.database import User

_LOGGER = logging.getLogger(__name__)


class UserManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user_model = User

    def create_user(self, params: dict) -> User:
        def _rollback(vo: User):
            _LOGGER.info(
                f"[create_user._rollback] Delete user: {vo.user_id} ({vo.name})"
            )
            vo.delete()

        if timezone := params.get("timezone"):
            self._check_timezone(timezone)

        if email := params.get("email"):
            self._check_email_format(email)

        if params["auth_type"] == "EXTERNAL":
            params["password"] = None

        else:
            if password := params.get("password"):
                self._check_password_format(password)
            else:
                raise ERROR_REQUIRED_PARAMETER(key="password")

            hashed_pw = PasswordCipher().hashpw(password)
            params["password"] = hashed_pw

        params["name"] = params["name"].strip()

        if email := params.get("email"):
            params["email"] = email.strip()

        user_vo = self.user_model.create(params)

        self.transaction.add_rollback(_rollback, user_vo)

        return user_vo

    def update_user_by_vo(self, params: dict, user_vo: User) -> User:
        def _rollback(old_data):
            _LOGGER.info(
                f'[update_user_by_vo._rollback] Revert Data: {old_data["user_id"]}'
            )
            user_vo.update(old_data)

        if timezone := params.get("timezone"):
            self._check_timezone(timezone)

        if email := params.get("email"):
            self._check_email_format(email)

        required_actions = list(user_vo.required_actions)
        is_change_required_actions = False

        if new_password := params.get("password"):
            if PasswordCipher().checkpw(new_password, user_vo.password):
                raise ERROR_PASSWORD_NOT_CHANGED(user_id=user_vo.user_id)

            self._check_password_format(params["password"])
            hashed_pw = PasswordCipher().hashpw(params["password"])
            params["password"] = hashed_pw

            if "UPDATE_PASSWORD" in required_actions:
                required_actions.remove("UPDATE_PASSWORD")
                is_change_required_actions = True

        if is_change_required_actions:
            params["required_actions"] = required_actions

        self.transaction.add_rollback(_rollback, user_vo.to_dict())

        return user_vo.update(params)

    @staticmethod
    def delete_user_by_vo(user_vo: User) -> None:
        rb_mgr = RoleBindingManager()
        user_group_mgr = UserGroupManager()
        project_mgr = ProjectManager()
        workspace_group_mgr = WorkspaceGroupManager()

        # Delete role bindings
        rb_vos = rb_mgr.filter_role_bindings(
            user_id=user_vo.user_id, domain_id=user_vo.domain_id
        )
        for rb_vo in rb_vos:
            rb_mgr.delete_role_binding_by_vo(rb_vo)

        # Delete user groups
        user_group_vos = user_group_mgr.filter_user_groups(
            users=user_vo.user_id, domain_id=user_vo.domain_id
        )
        for user_group_vo in user_group_vos:
            users = user_group_vo.users
            users.remove(user_vo.user_id)
            user_group_mgr.update_user_group_by_vo(
                {"users": users}, user_group_vo=user_group_vo
            )

        # Delete projects
        project_vos = project_mgr.filter_projects(
            users=user_vo.user_id, domain_id=user_vo.domain_id
        )
        for project_vo in project_vos:
            users = project_vo.users
            users.remove(user_vo.user_id)
            project_mgr.update_project_by_vo({"users": users}, project_vo=project_vo)

        # Delete workspace groups
        workspace_group_vos = workspace_group_mgr.filter_workspace_groups(
            users__user_id=user_vo.user_id, domain_id=user_vo.domain_id
        )

        for workspace_group_vo in workspace_group_vos:
            workspace_group_dict = workspace_group_vo.to_mongo().to_dict()
            users = workspace_group_dict.get("users", [])

            if users:
                updated_users = [
                    user for user in users if user.get("user_id") != user_vo.user_id
                ]

                if len(updated_users) != len(users):
                    workspace_group_mgr.update_workspace_group_by_vo(
                        {"users": updated_users}, workspace_group_vo=workspace_group_vo
                    )

        user_vo.delete()

    def get_user(self, user_id: str, domain_id: str) -> User:
        return self.user_model.get(user_id=user_id, domain_id=domain_id)

    def filter_users(self, **conditions) -> QuerySet:
        return self.user_model.filter(**conditions)

    def list_users(self, query: dict) -> Tuple[QuerySet, int]:
        return self.user_model.query(**query)

    def stat_users(self, query: dict) -> dict:
        return self.user_model.stat(**query)

    @staticmethod
    def _check_user_id_format(user_id: str) -> None:
        rule = r"[^@]+@[^@]+\.[^@]+"
        if not re.match(rule, user_id):
            raise ERROR_INCORRECT_USER_ID_FORMAT(rule="Email format required.")

    @staticmethod
    def _check_email_format(email: str) -> None:
        rule = r"[^@]+@[^@]+\.[^@]+"
        if not re.match(rule, email):
            raise ERROR_INCORRECT_EMAIL_FORMAT(rule="Email format required.")

    @staticmethod
    def _check_password_format(password: str) -> None:
        if len(password) < 8:
            raise ERROR_INCORRECT_PASSWORD_FORMAT(rule="At least 9 characters long.")
        elif not re.search("[a-z]", password):
            raise ERROR_INCORRECT_PASSWORD_FORMAT(
                rule="Contains at least one lowercase character"
            )
        elif not re.search("[A-Z]", password):
            raise ERROR_INCORRECT_PASSWORD_FORMAT(
                rule="Contains at least one uppercase character"
            )
        elif not re.search("[0-9]", password):
            raise ERROR_INCORRECT_PASSWORD_FORMAT(rule="Contains at least one number")

    @staticmethod
    def _check_timezone(timezone):
        if timezone not in pytz.all_timezones:
            raise ERROR_INVALID_PARAMETER(key="timezone", reason="Timezone is invalid.")



import logging
from typing import Union

from spaceone.core.manager import BaseManager
from spaceone.core.error import ERROR_NOT_FOUND

from spaceone.identity.error.custom import ERROR_MANAGED_RESOURCE_CAN_NOT_BE_MODIFIED
from spaceone.identity.model.service_account.database import ServiceAccount
from spaceone.identity.model.trusted_account.database import TrustedAccount
from spaceone.identity.model.project.database import Project
from spaceone.identity.model.project_group.database import ProjectGroup
from spaceone.identity.model.workspace.database import Workspace

_LOGGER = logging.getLogger(__name__)


class ResourceManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trusted_account_model = TrustedAccount

    def check_is_managed_resource_by_trusted_account(
        self,
        resource_vo: Union[ServiceAccount, Project, ProjectGroup, Workspace],
    ) -> None:
        try:
            if resource_vo.is_managed:
                trusted_account_vo = self.trusted_account_model.get(
                    trusted_account_id=resource_vo.trusted_account_id,
                    domain_id=resource_vo.domain_id,
                )
                if trusted_account_vo.schedule.get("state") == "ENABLED":
                    raise ERROR_MANAGED_RESOURCE_CAN_NOT_BE_MODIFIED()
        except ERROR_NOT_FOUND:
            _LOGGER.debug(
                f"[check_is_managed_resource] TrustedAccount not found. (trusted_account_id={resource_vo.trusted_account_id})"
            )
            return



import logging
from typing import Union

from spaceone.core.service import BaseService, transaction
from spaceone.core.service.utils import convert_model

from spaceone.identity.plugin.external_auth.model.external_auth_request import (
    ExternalAuthAuthorizeRequest,
    ExternalAuthInitRequest,
)
from spaceone.identity.plugin.external_auth.model.external_auth_response import (
    PluginResponse,
    UserResponse,
)

_LOGGER = logging.getLogger(__name__)


class ExternalAuthService(BaseService):
    resource = "ExternalAuth"

    @transaction
    @convert_model
    def init(self, params: ExternalAuthInitRequest) -> Union[PluginResponse, dict]:
        """Init external auth data

        Args:
            params (ExternalAuthInitRequest): {
                'options': 'dict',          # Required
                'domain_id': 'str'          # Required
            }

        Returns:
            Generator[PluginResponse, None, None]
            {
                'metadata': 'dict'
            }
        """

        func = self.get_plugin_method("init")
        response = func(params.dict())
        return PluginResponse(**response)

    @transaction
    @convert_model
    def authorize(
        self, params: ExternalAuthAuthorizeRequest
    ) -> Union[UserResponse, dict]:
        """Authorize external auth data

        Args:
            params (ExternalAuthAuthorizeRequest): {
                'options': 'dict',          # Required
                'schema_id': 'str',
                'secret_data': 'dict',      # Required
                'credentials': 'dict',      # Required
                'domain_id': 'str'          # Required
            }

        Returns:
            UserResponse: {
                'state': 'str',
                'user_id': 'str',
                'name': 'str',
                'email': 'str',
                'mobile': 'str',
                'group': 'str',
            }
        """

        func = self.get_plugin_method("authorize")
        response = func(params.dict())
        return UserResponse(**response)



import logging
from typing import Tuple, List
from mongoengine import QuerySet

from spaceone.core import cache
from spaceone.core.manager import BaseManager

from spaceone.identity.model.project.database import Project
from spaceone.identity.error.error_project import *
from spaceone.identity.manager.service_account_manager import ServiceAccountManager

_LOGGER = logging.getLogger(__name__)


class ProjectManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.project_model = Project

    def create_project(self, params: dict) -> Project:
        def _rollback(vo: Project):
            _LOGGER.info(
                f"[create_project._rollback] Delete project: {vo.name} ({vo.project_id})"
            )
            vo.delete()

        project_vo = self.project_model.create(params)
        self.transaction.add_rollback(_rollback, project_vo)

        return project_vo

    def update_project_by_vo(self, params: dict, project_vo: Project) -> Project:
        def _rollback(old_data):
            _LOGGER.info(
                f'[update_project._rollback] Revert Data: {old_data["name"]} ({old_data["project_id"]})'
            )
            project_vo.update(old_data)

        self.transaction.add_rollback(_rollback, project_vo.to_dict())

        return project_vo.update(params)

    @staticmethod
    def delete_project_by_vo(project_vo: Project) -> None:
        service_account_mgr = ServiceAccountManager()
        service_account_vos = service_account_mgr.filter_service_accounts(
            project_id=project_vo.project_id
        )
        for service_account_vo in service_account_vos:
            raise ERROR_RELATED_SERVICE_ACCOUNT_EXIST(
                service_account_id=service_account_vo.service_account_id
            )

        project_vo.delete()

    def get_project(
        self,
        project_id: str,
        domain_id: str,
        workspace_id: str = None,
        user_projects: List[str] = None,
    ) -> Project:
        if user_projects and project_id not in user_projects:
            raise ERROR_PERMISSION_DENIED()

        conditions = {
            "project_id": project_id,
            "domain_id": domain_id,
        }

        if workspace_id:
            conditions.update({"workspace_id": workspace_id})

        return self.project_model.get(**conditions)

    def filter_projects(self, **conditions) -> QuerySet:
        return self.project_model.filter(**conditions)

    def list_projects(self, query: dict) -> Tuple[QuerySet, int]:
        return self.project_model.query(**query)

    def stat_projects(self, query: dict) -> dict:
        return self.project_model.stat(**query)



from mongoengine import (
    DateTimeField,
    DictField,
    EmbeddedDocument,
    EmbeddedDocumentField,
    IntField,
    ListField,
    StringField,
)
from spaceone.core.model.mongo_model import MongoModel


class WorkspaceGroupUser(EmbeddedDocument):
    user_id = StringField(max_length=40, required=True)
    role_id = StringField(max_length=40, required=True)
    role_type = StringField(
        max_length=20, choices=("WORKSPACE_OWNER", "WORKSPACE_MEMBER")
    )


class WorkspaceGroup(MongoModel):
    workspace_group_id = StringField(max_length=40, generate_id="wg", unique=True)
    name = StringField(max_length=255, unique_with="domain_id")
    workspace_count = IntField(default=None)
    users = ListField(
        EmbeddedDocumentField(WorkspaceGroupUser), default=None, null=True
    )
    tags = DictField(default=None)
    created_by = StringField(max_length=255)
    updated_by = StringField(max_length=255)
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        "updatable_fields": [
            "name",
            "workspace_count",
            "users",
            "tags",
            "updated_by",
            "updated_at",
        ],
        "minimal_fields": [
            "workspace_group_id",
            "name",
        ],
        "change_query_keys": {},
        "ordering": ["name"],
        "indexes": [
            "name",
        ],
    }



from typing import Union

from pydantic import BaseModel

__all__ = [
    "UserProfileUpdateRequest",
    "UserProfileVerifyEmailRequest",
    "UserProfileConfirmEmailRequest",
    "UserProfileResetPasswordRequest",
    "UserProfileEnableMFARequest",
    "UserProfileDisableMFARequest",
    "UserProfileConfirmMFARequest",
    "UserProfileGetRequest",
    "UserProfileGetWorkspacesRequest",
    "UserProfileGetWorkspaceGroupsRequest",
]


class UserProfileUpdateRequest(BaseModel):
    user_id: str
    password: Union[str, None] = None
    name: Union[str, None] = None
    email: Union[str, None] = None
    language: Union[str, None] = None
    timezone: Union[str, None] = None
    tags: Union[dict, None] = None
    domain_id: str


class UserProfileVerifyEmailRequest(BaseModel):
    user_id: str
    email: Union[str, None] = None
    domain_id: str


class UserProfileConfirmEmailRequest(BaseModel):
    user_id: str
    verify_code: str
    domain_id: str


class UserProfileResetPasswordRequest(BaseModel):
    user_id: str
    domain_id: str


class UserProfileEnableMFARequest(BaseModel):
    user_id: str
    mfa_type: str
    options: dict
    domain_id: str


class UserProfileDisableMFARequest(BaseModel):
    user_id: str
    domain_id: str


class UserProfileConfirmMFARequest(BaseModel):
    user_id: str
    verify_code: str
    domain_id: str


class UserProfileGetRequest(BaseModel):
    user_id: str
    domain_id: str


class UserProfileGetWorkspacesRequest(BaseModel):
    workspace_group_id: Union[str, None] = None
    user_id: str
    domain_id: str


class UserProfileGetWorkspaceGroupsRequest(BaseModel):
    user_id: str
    domain_id: str



import logging
from typing import Union

from spaceone.core.service import *
from spaceone.core.service.utils import *

from spaceone.identity.error.error_external_auth import ERROR_REQUIRED_FIELDS
from spaceone.identity.manager.domain_manager import DomainManager
from spaceone.identity.manager.external_auth_manager import ExternalAuthManager
from spaceone.identity.model.external_auth.request import *
from spaceone.identity.model.external_auth.response import *

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class ExternalAuthService(BaseService):
    resource = "ExternalAuth"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.external_auth_mgr = ExternalAuthManager()

    @transaction(permission="identity:ExternalAuth.write", role_types=["DOMAIN_ADMIN"])
    @convert_model
    def set(self, params: ExternalAuthSetRequest) -> Union[ExternalAuthResponse, dict]:
        """Set external auth info
        Args:
            params (dict): {
                'domain_id': 'str',    # injected from auth (required)
                'plugin_info': 'dict'
            }
        Returns:
            ExternalAuthResponse:
        """

        domain_mgr = DomainManager()
        domain_vo = domain_mgr.get_domain(params.domain_id)

        self._validate_params(params)

        external_auth_vo = self.external_auth_mgr.set_external_auth(
            params.dict(), domain_vo
        )

        return ExternalAuthResponse(**external_auth_vo.to_dict())

    @transaction(permission="identity:ExternalAuth.write", role_types=["DOMAIN_ADMIN"])
    @convert_model
    def unset(
        self, params: ExternalAuthUnsetRequest
    ) -> Union[ExternalAuthResponse, dict]:
        """Unset external auth info
        Args:
            params (dict): {
                'domain_id': 'str'  # injected from auth (required)
            }
        Returns:
            ExternalAuthResponse:
        """

        external_auth_vo = self.external_auth_mgr.get_external_auth(params.domain_id)
        self.external_auth_mgr.delete_external_auth_by_vo(external_auth_vo)

        return {"domain_id": params.domain_id, "state": "DISABLED"}

    @transaction(permission="identity:ExternalAuth.read", role_types=["DOMAIN_ADMIN"])
    @convert_model
    def get(self, params: ExternalAuthGetRequest) -> Union[ExternalAuthResponse, dict]:
        """Get external auth info
        Args:
            params (dict): {
                'domain_id': 'str' # injected from auth (required)
            }
        Returns:
            ExternalAuthResponse:
        """

        external_auth_vos = self.external_auth_mgr.filter_external_auth(
            domain_id=params.domain_id
        )
        if external_auth_vos.count() > 0:
            external_auth_vo = external_auth_vos[0]
            return ExternalAuthResponse(**external_auth_vo.to_dict())
        else:
            return {"domain_id": params.domain_id, "state": "DISABLED"}

    @staticmethod
    def _validate_params(params):
        plugin_info = params.dict().get("plugin_info", {})
        options = plugin_info.get("options", {})

        protocol = options.get("protocol", "")
        identity_provider = options.get("identity_provider", "")

        if (not protocol) or (not identity_provider):
            if (not protocol) and (not identity_provider):
                _LOGGER.error(
                    "[_validate_params] ERROR_REQUIRED_FIELDS: options.protocol, options.identity_provider"
                )
                raise ERROR_REQUIRED_FIELDS(
                    field="options.protocol, options.identity_provider"
                )
            elif not protocol:
                _LOGGER.error(
                    "[_validate_params] ERROR_REQUIRED_FIELDS: options.protocol"
                )
                raise ERROR_REQUIRED_FIELDS(field="options.protocol")
            elif not identity_provider:
                _LOGGER.error(
                    "[_validate_params] ERROR_REQUIRED_FIELDS: options.identity_provider"
                )
                raise ERROR_REQUIRED_FIELDS(field="options.identity_provider")



from pydantic import BaseModel
from spaceone.identity.model.domain.request import AdminUser

__all__ = [
    "SystemInitRequest",
]


class SystemInitRequest(BaseModel):
    admin: AdminUser
    force: bool = False



from spaceone.core.error import *


class ERROR_ROLE_IN_USED(ERROR_INVALID_ARGUMENT):
    _message = (
        "Role is used. (role_binding_id = {role_binding_id}, user_id = {user_id})"
    )


class ERROR_NOT_ALLOWED_ROLE_TYPE(ERROR_INVALID_ARGUMENT):
    _message = (
        "Role type is not allowed. (request_role_id = {request_role_id}, "
        "request_role_type = {request_role_type}, supported_role_type = {supported_role_type})"
    )


class ERROR_NOT_ALLOWED_USER_STATE(ERROR_INVALID_ARGUMENT):
    _message = "User state is not allowed. (user_id = {user_id}, state = {state})"


class ERROR_DUPLICATED_ROLE_BINDING(ERROR_INVALID_ARGUMENT):
    _message = (
        "Role type is duplicated. (role_type = {role_type}), "
        "Only one {role_type} is allowed per user in the domain."
    )


class ERROR_DUPLICATED_WORKSPACE_ROLE_BINDING(ERROR_INVALID_ARGUMENT):
    _message = "Only one role binding is allowed per user in then same workspace. (choices = {allowed_role_type})"


class ERROR_NOT_ALLOWED_TO_UPDATE_OR_DELETE_ROLE_BY_SELF(ERROR_INVALID_ARGUMENT):
    _message = "You are not allowed to update or delete your own role."


class ERROR_LAST_WORKSPACE_OWNER_CANNOT_DELETE(ERROR_INVALID_ARGUMENT):
    _message = "Last workspace owner cannot be deleted."


class ERROR_LAST_DOMAIN_ADMIN_CANNOT_DELETE(ERROR_INVALID_ARGUMENT):
    _message = "Last domain admin cannot be deleted."



import logging
from datetime import datetime

from spaceone.core.error import ERROR_CONFIGURATION
from spaceone.core import config
from spaceone.core import utils
from spaceone.core.locator import Locator
from spaceone.core.scheduler import HourlyScheduler

_LOGGER = logging.getLogger(__name__)


class TrustedAccountSyncScheduler(HourlyScheduler):
    def __init__(self, queue, interval, minute=":00"):
        super().__init__(queue, interval, minute)
        self.locator = Locator()
        self._init_config()

    def _init_config(self):
        self._token = config.get_global("TOKEN")
        if self._token is None:
            raise ERROR_CONFIGURATION(key="TOKEN")

    def create_task(self) -> list:
        tasks = []
        tasks.extend(self._create_trusted_account_sync_task())
        return tasks

    def _create_trusted_account_sync_task(self):
        current_hour = datetime.utcnow().hour
        stp = {
            "name": "trusted_account_sync_schedule",
            "version": "v1",
            "executionEngine": "BaseWorker",
            "stages": [
                {
                    "locator": "SERVICE",
                    "name": "JobService",
                    "metadata": {"token": self._token},
                    "method": "create_jobs_by_trusted_account",
                    "params": {"params": {"current_hour": current_hour}},
                }
            ],
        }
        print(
            f"{utils.datetime_to_iso8601(datetime.utcnow())} [INFO] [create_task] create_jobs_by_trusted_account => START"
        )
        return [stp]



import logging

from spaceone.core import utils
from spaceone.core.manager import BaseManager
from spaceone.identity.lib.key_generator import KeyGenerator
from spaceone.identity.manager.domain_secret_manager import DomainSecretManager

_LOGGER = logging.getLogger(__name__)


class SystemManager(BaseManager):
    @staticmethod
    def get_root_domain_id():
        return "domain-root"

    @staticmethod
    def create_system_token(root_domain_id: str, admin_user_id: str) -> str:
        domain_secret_mgr = DomainSecretManager()
        prv_jwk = domain_secret_mgr.get_domain_private_key(root_domain_id)

        key_gen = KeyGenerator(prv_jwk, root_domain_id, "SYSTEM", admin_user_id)
        return key_gen.generate_token("SYSTEM_TOKEN")



from spaceone.core.error import *


class ERROR_AUTHENTICATION_FAILURE(ERROR_AUTHENTICATE_FAILURE):
    _message = 'A user "{user_id}" is not exist or password is not correct.'


class ERROR_NOT_AUTHENTICATED(ERROR_AUTHENTICATE_FAILURE):
    _message = "A user is not authenticated."


class ERROR_INVALID_CREDENTIALS(ERROR_AUTHENTICATE_FAILURE):
    _message = "Invalid credentials provided."


class ERROR_AUTHENTICATION_FAILURE_PLUGIN(ERROR_INTERNAL_API):
    _message = "External plugin authentication exception. (reason = {message})"


class ERROR_INVALID_GRANT_TYPE(ERROR_INVALID_ARGUMENT):
    _message = "Invalid grant type. (grant_type = {grant_type})"


class ERROR_UPDATE_PASSWORD_REQUIRED(ERROR_INVALID_ARGUMENT):
    _message = "Password reset is required.(user_id = {user_id})"



import logging
from typing import Dict, List, Tuple

from mongoengine import QuerySet
from spaceone.core.error import ERROR_INVALID_PARAMETER, ERROR_NOT_FOUND
from spaceone.core.manager import BaseManager

from spaceone.identity.error import (
    ERROR_USER_EXIST_IN_WORKSPACE_GROUP,
    ERROR_WORKSPACE_EXIST_IN_WORKSPACE_GROUP,
)
from spaceone.identity.manager.role_binding_manager import RoleBindingManager
from spaceone.identity.manager.workspace_manager import WorkspaceManager
from spaceone.identity.model.workspace_group.database import WorkspaceGroup

_LOGGER = logging.getLogger(__name__)


class WorkspaceGroupManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.workspace_group_model = WorkspaceGroup
        self.workspace_manager = WorkspaceManager()
        self.rb_mgr = RoleBindingManager()

    def create_workspace_group(self, params: dict) -> WorkspaceGroup:
        def _rollback(vo: WorkspaceGroup):
            _LOGGER.info(
                f"[create_workspace_group._rollback] Delete workspace group : {vo.workspace_group_id} ({vo.name})"
            )
            vo.delete()

        workspace_group_vo = self.workspace_group_model.create(params)
        self.transaction.add_rollback(_rollback, workspace_group_vo)

        return workspace_group_vo

    def update_workspace_group_by_vo(
        self, params: dict, workspace_group_vo: WorkspaceGroup
    ) -> WorkspaceGroup:
        def _rollback(old_data):
            _LOGGER.info(
                f"[update_workspace_group._rollback] Revert Data : {old_data['workspace_group_id']} ({old_data['name']})"
            )
            workspace_group_vo.update(old_data)

        self.transaction.add_rollback(_rollback, workspace_group_vo.to_dict())

        return workspace_group_vo.update(params)

    def delete_workspace_group_by_vo(self, workspace_group_vo: WorkspaceGroup) -> None:
        workspace_vos = self.workspace_manager.filter_workspaces(
            workspace_group_id=workspace_group_vo.workspace_group_id,
            domain_id=workspace_group_vo.domain_id,
        )
        self._check_existence_of_workspace_or_user(workspace_vos, workspace_group_vo)

        workspace_group_vo.delete()

    def get_workspace_group(
        self, domain_id: str, workspace_group_id: str, user_id: str = None
    ) -> WorkspaceGroup:
        conditions = {
            "domain_id": domain_id,
            "workspace_group_id": workspace_group_id,
        }

        if user_id:
            conditions["users__user_id"] = user_id

        return self.workspace_group_model.get(**conditions)

    def filter_workspace_groups(self, **conditions) -> QuerySet:
        return self.workspace_group_model.filter(**conditions)

    def list_workspace_groups(self, query: dict) -> Tuple[QuerySet, int]:
        return self.workspace_group_model.query(**query)

    def stat_workspace_group(self, query: dict) -> dict:
        return self.workspace_group_model.stat(**query)

    def get_workspace_group_with_users(
        self, domain_id: str, workspace_group_id: str
    ) -> Tuple[WorkspaceGroup, List[str]]:
        workspace_group_vo = self.get_workspace_group(domain_id, workspace_group_id)

        workspace_group_user_ids = []
        if workspace_group_vo.users:
            old_users, new_users = self.get_unique_user_ids(
                domain_id, workspace_group_id, workspace_group_vo.users
            )
            workspace_group_user_ids = old_users + new_users

        return workspace_group_vo, workspace_group_user_ids

    @staticmethod
    def _check_existence_of_workspace_or_user(
        workspace_vos: QuerySet, workspace_group_vo: WorkspaceGroup
    ) -> None:
        for workspace_vo in workspace_vos:
            if workspace_vo:
                _LOGGER.error(
                    f"Workspace exists in Workspace Group. ({workspace_vo.workspace_id}, {workspace_group_vo.workspace_group_id})"
                    "Remove the workspace from the workspace group before deleting the workspace group."
                )
                raise ERROR_WORKSPACE_EXIST_IN_WORKSPACE_GROUP(
                    workspace_id=workspace_vo.workspace_id,
                    workspace_group_id=workspace_group_vo.workspace_group_id,
                )

        if users := workspace_group_vo.users:
            for user in users:
                _LOGGER.error(
                    f"User exists in Workspace Group. ({user['user_id']}, {workspace_group_vo.workspace_group_id})"
                    "Remove the user from the workspace group before deleting the workspace group."
                )
                raise ERROR_USER_EXIST_IN_WORKSPACE_GROUP(
                    user_id=user["user_id"],
                    workspace_group_id=workspace_group_vo.workspace_group_id,
                )

    @staticmethod
    def check_user_id_in_users(user_id: str, workspace_group_vo: WorkspaceGroup):
        return any(
            workspace_group_user_id == user_id
            for workspace_group_user_id in workspace_group_vo["users"]
        )

    def get_unique_user_ids(
        self,
        domain_id: str,
        workspace_group_id: str,
        new_users_info_list: List[Dict[str, str]],
    ) -> Tuple[List[str], List[str]]:
        workspace_group_vo = self.get_workspace_group(domain_id, workspace_group_id)

        old_user_ids = []
        if users := workspace_group_vo.users:
            unique_user_ids = {user["user_id"] for user in users}
            old_user_ids = list(unique_user_ids)

        new_user_ids = []
        for new_user_info in new_users_info_list:
            unique_user_ids = {new_user_info["user_id"]}
            new_user_ids = list(unique_user_ids)

        return old_user_ids, new_user_ids

    @staticmethod
    def check_new_users_already_in_workspace_group(
        old_user_ids: List[str], new_user_ids: List[str]
    ) -> None:
        if set(old_user_ids) & set(new_user_ids):
            _LOGGER.error(
                f"Users {new_user_ids} is already in workspace group or not registered."
            )
            raise ERROR_INVALID_PARAMETER(
                key="users",
                reason=f"User {new_user_ids} is already in the workspace group or not registered.",
            )

    @staticmethod
    def check_user_ids_exist_in_workspace_group(
        old_user_ids: List[str], user_ids: List[str]
    ) -> None:
        for user_id in user_ids:
            if user_id not in old_user_ids:
                _LOGGER.error(f"User ID {user_id} is not in workspace group.")
                raise ERROR_NOT_FOUND(key="user_id", value=user_id)



import logging
from datetime import datetime
from mongoengine import QuerySet
from typing import Dict, List, Union, Any

from spaceone.core.error import ERROR_INVALID_PARAMETER, ERROR_NOT_FOUND
from spaceone.core.service import (
    BaseService,
    authentication_handler,
    authorization_handler,
    event_handler,
    mutation_handler,
    transaction,
)
from spaceone.core.service.utils import (
    append_keyword_filter,
    append_query_filter,
    convert_model,
)

from spaceone.identity.error.error_role import (
    ERROR_NOT_ALLOWED_ROLE_TYPE,
    ERROR_NOT_ALLOWED_USER_STATE,
)
from spaceone.identity.manager.role_binding_manager import RoleBindingManager
from spaceone.identity.manager.role_manager import RoleManager
from spaceone.identity.manager.user_manager import UserManager
from spaceone.identity.manager.workspace_group_manager import WorkspaceGroupManager
from spaceone.identity.manager.workspace_group_user_manager import (
    WorkspaceGroupUserManager,
)
from spaceone.identity.manager.workspace_manager import WorkspaceManager
from spaceone.identity.model import WorkspaceGroup
from spaceone.identity.model.workspace_group.database import WorkspaceGroupUser
from spaceone.identity.model.workspace_group.request import (
    WorkspaceGroupAddUsersRequest,
    WorkspaceGroupCreateRequest,
    WorkspaceGroupDeleteRequest,
    WorkspaceGroupGetRequest,
    WorkspaceGroupRemoveUsersRequest,
    WorkspaceGroupSearchQueryRequest,
    WorkspaceGroupStatQueryRequest,
    WorkspaceGroupUpdateRequest,
    WorkspaceGroupUpdateRoleRequest,
)
from spaceone.identity.model.workspace_group.response import (
    WorkspaceGroupResponse,
    WorkspaceGroupsResponse,
)
from spaceone.identity.model.workspace_group_user.request import (
    WorkspaceGroupUserAddRequest,
    WorkspaceGroupUserRemoveRequest,
    WorkspaceGroupUserUpdateRoleRequest,
)
from spaceone.identity.service.role_binding_service import RoleBindingService

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class WorkspaceGroupService(BaseService):
    resource = "WorkspaceGroup"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.workspace_group_mgr = WorkspaceGroupManager()
        self.workspace_group_user_mgr = WorkspaceGroupUserManager()
        self.workspace_mgr = WorkspaceManager()
        self.user_mgr = UserManager()
        self.role_mgr = RoleManager()
        self.rb_mgr = RoleBindingManager()
        self.rb_svc = RoleBindingService()

    @transaction(
        permission="identity:WorkspaceGroup.write", role_types=["DOMAIN_ADMIN"]
    )
    @convert_model
    def create(
        self, params: WorkspaceGroupCreateRequest
    ) -> Union[WorkspaceGroupResponse, dict]:
        """Create workspace group
        Args:
            params (WorkspaceGroupCreateRequest): {
                'name': 'str',                  # required
                'tags': 'dict',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str',             # injected from auth (required)
            }
        Returns:
            WorkspaceGroupResponse:
        """
        params_data = params.dict(exclude_unset=True)
        params_data["created_by"] = self.transaction.get_meta("authorization.audience")
        params_data["workspace_count"] = 0
        workspace_group_vo = self.workspace_group_mgr.create_workspace_group(
            params_data
        )
        return WorkspaceGroupResponse(**workspace_group_vo.to_dict())

    @transaction(
        permission="identity:WorkspaceGroup.write", role_types=["DOMAIN_ADMIN"]
    )
    @convert_model
    def update(
        self, params: WorkspaceGroupUpdateRequest
    ) -> Union[WorkspaceGroupResponse, dict]:
        """Update workspace group name and tags
        Args:
            params (WorkspaceGroupUpdateRequest): {
                'workspace_group_id': 'str',    # required
                'name': 'str',
                'tags': 'dict',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str',             # injected from auth (required)
        Returns:
            WorkspaceGroupResponse:
        """
        workspace_group_vo = self.workspace_group_mgr.get_workspace_group(
            params.domain_id, params.workspace_group_id
        )

        params_data = params.dict(exclude_unset=True)
        params_data["updated_by"] = self.transaction.get_meta("authorization.audience")
        params_data["updated_at"] = datetime.utcnow()

        workspace_group_vo = self.workspace_group_mgr.update_workspace_group_by_vo(
            params_data, workspace_group_vo
        )

        return WorkspaceGroupResponse(**workspace_group_vo.to_dict())

    @transaction(
        permission="identity:WorkspaceGroup.write", role_types=["DOMAIN_ADMIN"]
    )
    @convert_model
    def delete(self, params: WorkspaceGroupDeleteRequest) -> None:
        """Delete workspace group
        Args:
            params (WorkspaceGroupDeleteRequest): {
                'workspace_group_id': 'str',    # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str',             # injected from auth (required)
            }
        Returns:
            None
        """
        workspace_group_vo = self.workspace_group_mgr.get_workspace_group(
            params.domain_id, params.workspace_group_id
        )

        self.workspace_group_mgr.delete_workspace_group_by_vo(workspace_group_vo)

    @transaction(
        permission="identity:WorkspaceGroup.write", role_types=["DOMAIN_ADMIN"]
    )
    @convert_model
    def add_users(
        self, params: WorkspaceGroupAddUsersRequest
    ) -> Union[WorkspaceGroupResponse, dict]:
        """Add users to workspace group
        Args:
            params (WorkspaceGroupAddUsersRequest): {
                'workspace_group_id': 'str',      # required
                'users': [                        # required
                    {
                        'user_id': 'str',         # required
                        'role_id': 'str',         # required
                    }
                ],
                'workspace_id': 'str',            # injected from auth
                'domain_id': 'str',               # injected from auth (required)
            }
        Returns:
           WorkspaceGroupResponse:
        """
        return self.process_add_users(params, role_type="DOMAIN_ADMIN")

    @transaction(
        permission="identity:WorkspaceGroup.write", role_types=["DOMAIN_ADMIN"]
    )
    @convert_model
    def remove_users(
        self, params: WorkspaceGroupRemoveUsersRequest
    ) -> Union[WorkspaceGroupResponse, dict]:
        """Remove users from workspace group
        Args:
            params (WorkspaceGroupRemoveUsersRequest): {
                'workspace_group_id': 'str',         # required
                'users': [
                    'user_id': 'str'
                ],                                   # required
                'workspace_id': 'str',               # injected from auth
                'domain_id': 'str',                  # injected from auth (required)
            }
        Returns:
            WorkspaceGroupResponse:
        """
        return self.process_remove_users(params, role_type="DOMAIN_ADMIN")

    @transaction(
        permission="identity:WorkspaceGroup.write", role_types=["DOMAIN_ADMIN"]
    )
    @convert_model
    def update_role(
        self, params: WorkspaceGroupUpdateRoleRequest
    ) -> Union[WorkspaceGroupResponse, dict]:
        """Update workspace group's user's role

        Args:
            params (WorkspaceGroupUpdateRoleRequest): {
                'workspace_group_id': 'str',        # required
                'user_id': 'str',                   # required
                'role_id': 'str',                   # required
                'workspace_id': 'str',              # injected from auth
                'domain_id': 'str',                 # injected from auth (required)
            }
        Returns:
            WorkspaceGroupResponse:
        """
        return self.process_update_role(params, role_type="DOMAIN_ADMIN")

    @transaction(permission="identity:WorkspaceGroup.read", role_types=["DOMAIN_ADMIN"])
    @convert_model
    def get(
        self, params: WorkspaceGroupGetRequest
    ) -> Union[WorkspaceGroupResponse, dict]:
        """Get workspace group

        Args:
            params (WorkspaceGroupGetRequest): {
                'workspace_group_id': 'str', # required
                'workspace_id': 'str',       # injected from auth
                'domain_id': 'str',          # injected from auth (required)
            }
        Returns:
            WorkspaceGroupResponse:
        """
        workspace_group_vo, workspace_group_user_ids = (
            self.workspace_group_mgr.get_workspace_group_with_users(
                params.domain_id, params.workspace_group_id
            )
        )
        workspace_group_info = self.add_user_name_and_state_to_users(
            workspace_group_user_ids, workspace_group_vo, params.domain_id
        )
        return WorkspaceGroupResponse(**workspace_group_info)

    @transaction(permission="identity:WorkspaceGroup.read", role_types=["DOMAIN_ADMIN"])
    @append_query_filter(["workspace_group_id", "name", "domain_id"])
    @append_keyword_filter(["workspace_group_id", "name"])
    @convert_model
    def list(
        self, params: WorkspaceGroupSearchQueryRequest
    ) -> Union[WorkspaceGroupsResponse, dict]:
        """List workspace groups
        Args:
            params (WorkspaceGroupSearchQueryRequest): {
                'query': 'dict (spaceone.api.core.v1.Query)',
                'workspace_group_id': 'str',         # required
                'name': 'str',
                'created_by': 'str',
                'updated_by': 'str',
                'workspace_id': 'str',               # injected from auth
                'domain_id': 'str',                  # injected from auth (required)
            }
        Returns:
            WorkspaceGroupsResponse:
        """
        query = params.query

        workspace_group_vos, total_count = (
            self.workspace_group_mgr.list_workspace_groups(query)
        )
        workspace_groups_info = self.get_workspace_groups_info(
            workspace_group_vos, params.domain_id
        )

        return WorkspaceGroupsResponse(
            results=workspace_groups_info, total_count=total_count
        )

    @transaction(permission="identity:WorkspaceGroup.read", role_types=["DOMAIN_ADMIN"])
    @append_query_filter(["workspace_id", "workspace_group_id", "domain_id"])
    @convert_model
    def stat(self, params: WorkspaceGroupStatQueryRequest) -> dict:
        """Stat workspace groups
        Args:
            params (WorkspaceGroupStatQueryRequest): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)', # required
                'workspace_id': 'str',             # injected from auth
                'domain_id': 'str',                # injected from auth (required)
            }
        Returns:
            dict: {
                'results': 'list',
                'total_count': 'int'
            }
        """
        query = params.query

        return self.workspace_group_mgr.stat_workspace_group(query)

    def process_add_users(
        self,
        params: Union[WorkspaceGroupAddUsersRequest, WorkspaceGroupUserAddRequest],
        role_type: str,
    ) -> WorkspaceGroupResponse:
        domain_id = params.domain_id
        workspace_group_id = params.workspace_group_id
        new_users_info_list = params.users

        old_user_ids, new_user_ids = self.workspace_group_mgr.get_unique_user_ids(
            domain_id, workspace_group_id, new_users_info_list
        )
        self.workspace_group_mgr.check_new_users_already_in_workspace_group(
            old_user_ids, new_user_ids
        )
        self.check_new_users_exist_in_domain(new_user_ids, domain_id)

        workspace_group_vo = self.workspace_group_mgr.get_workspace_group(
            domain_id, workspace_group_id
        )

        if role_type == "USER":
            user_id = params.user_id
            user_vo = self.user_mgr.get_user(user_id, domain_id)

            # Role type USER has two roles: DOMAIN_ADMIN and USER
            if user_vo.role_type == "USER":
                workspace_group_old_users_info = workspace_group_vo.users or []
                if workspace_group_old_users_info:
                    self.workspace_group_user_mgr.check_user_role_type(
                        workspace_group_old_users_info, user_id
                    )

        new_users_role_map = self.get_role_map(new_users_info_list, domain_id)
        workspace_group_workspace_ids = self.get_workspace_ids(
            workspace_group_id, domain_id
        )
        if workspace_group_workspace_ids:
            self.delete_workspace_users_role_binding(
                new_user_ids, workspace_group_workspace_ids, domain_id
            )

        users_info_list = self.get_users_info_list(
            new_users_info_list,
            new_users_role_map,
            workspace_group_workspace_ids,
            workspace_group_vo,
            domain_id,
        )
        params.users = users_info_list

        workspace_group_vo = self.workspace_group_mgr.update_workspace_group_by_vo(
            params.dict(exclude_unset=True), workspace_group_vo
        )

        workspace_group_user_ids: List[str] = old_user_ids + new_user_ids
        workspace_group_user_info = self.add_user_name_and_state_to_users(
            workspace_group_user_ids, workspace_group_vo, domain_id
        )

        return WorkspaceGroupResponse(**workspace_group_user_info)

    def process_remove_users(
        self,
        params: Union[
            WorkspaceGroupRemoveUsersRequest, WorkspaceGroupUserRemoveRequest
        ],
        role_type: str,
    ) -> WorkspaceGroupResponse:
        domain_id = params.domain_id
        workspace_group_id = params.workspace_group_id
        users = params.users

        old_user_ids, user_ids = self.workspace_group_mgr.get_unique_user_ids(
            domain_id, workspace_group_id, users
        )
        self.workspace_group_mgr.check_user_ids_exist_in_workspace_group(
            old_user_ids, user_ids
        )

        params_users_user_id = users[0].get("user_id")
        workspace_group_vo = self.workspace_group_mgr.get_workspace_group(
            domain_id, workspace_group_id, user_id=params_users_user_id
        )
        if not workspace_group_vo:
            ERROR_NOT_FOUND(key="workspace_group_id", value=workspace_group_id)

        workspace_group_dict = workspace_group_vo.to_mongo().to_dict()

        if role_type == "USER":
            user_id = params.user_id
            user_vo = self.user_mgr.get_user(user_id, domain_id)

            # Role type USER has two roles: DOMAIN_ADMIN and USER
            if user_vo.role_type == "USER":
                workspace_group_users = workspace_group_vo.users
                self.workspace_group_user_mgr.check_user_role_type(
                    workspace_group_users, user_id
                )

        old_users = workspace_group_dict["users"]
        updated_users = self.remove_users_from_workspace_group(
            user_ids, old_users, workspace_group_id, domain_id
        )
        params.users = updated_users

        workspace_group_vo = self.workspace_group_mgr.update_workspace_group_by_vo(
            params.dict(exclude_unset=True), workspace_group_vo
        )

        return WorkspaceGroupResponse(**workspace_group_vo.to_dict())

    def process_update_role(
        self,
        params: Union[
            WorkspaceGroupUpdateRoleRequest, WorkspaceGroupUserUpdateRoleRequest
        ],
        role_type: str,
    ) -> WorkspaceGroupResponse:
        workspace_group_id = params.workspace_group_id
        role_id = params.role_id
        domain_id = params.domain_id

        workspace_group_vo = self.workspace_group_mgr.get_workspace_group(
            domain_id, workspace_group_id
        )

        if role_type == "DOMAIN_ADMIN":
            target_user_id = params.user_id
        else:
            # when role_type is USER, user_id is injected from auth, so the target_user_id is
            # the target user_id to update role
            target_user_id = params.target_user_id
            user_id = params.user_id

            user_vo = self.user_mgr.get_user(user_id, domain_id)

            # Role type USER has two roles: DOMAIN_ADMIN and USER
            if user_vo.role_type == "USER":
                workspace_group_users = workspace_group_vo.users
                self.workspace_group_user_mgr.check_user_role_type(
                    workspace_group_users, user_id
                )

        target_user_vo = self.user_mgr.get_user(target_user_id, domain_id)
        target_user_state = target_user_vo.state
        self.check_user_state(target_user_id, target_user_state)

        role_vo = self.role_mgr.get_role(role_id, domain_id)
        role_type = role_vo.role_type
        self.check_role_type(role_type)

        self.update_user_role_of_workspace_group(
            role_id, role_type, target_user_id, workspace_group_id, domain_id
        )

        update_workspace_group_params = {"users": workspace_group_vo.users or []}
        for user_info in update_workspace_group_params.get("users", []):
            if user_info["user_id"] == target_user_id:
                user_info["role_id"] = role_id
                user_info["role_type"] = role_type
                break

        workspace_group_vo = self.workspace_group_mgr.update_workspace_group_by_vo(
            update_workspace_group_params, workspace_group_vo
        )

        return WorkspaceGroupResponse(**workspace_group_vo.to_dict())

    def check_new_users_exist_in_domain(
        self, new_users: List[str], domain_id: str
    ) -> None:
        user_vos = self.user_mgr.filter_users(user_id=new_users, domain_id=domain_id)
        if not user_vos.count() == len(new_users):
            raise ERROR_NOT_FOUND(key="user_id", value=new_users)

    def get_role_map(
        self, users: List[Dict[str, str]], domain_id: str
    ) -> Dict[str, str]:
        role_ids = list(set([user["role_id"] for user in users]))
        role_vos = self.role_mgr.filter_roles(
            role_id=role_ids,
            domain_id=domain_id,
            role_type=["WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
        )
        role_map = {role_vo.role_id: role_vo.role_type for role_vo in role_vos}

        for role_id in role_ids:
            if role_id not in role_map:
                raise ERROR_INVALID_PARAMETER(
                    key=role_id, reason=f"{role_id} is invalid."
                )

        return role_map

    def get_workspace_ids(self, workspace_group_id: str, domain_id: str) -> List[str]:
        workspace_vos = self.workspace_mgr.filter_workspaces(
            workspace_group_id=workspace_group_id, domain_id=domain_id
        )
        workspace_ids = [workspace_vo.workspace_id for workspace_vo in workspace_vos]

        return workspace_ids

    def delete_workspace_users_role_binding(
        self,
        new_users: List[str],
        workspace_group_workspace_ids: List[str],
        domain_id: str,
    ) -> None:
        rb_vos = self.rb_mgr.filter_role_bindings(
            user_id=new_users,
            workspace_id=workspace_group_workspace_ids,
            domain_id=domain_id,
        )
        for rb_vo in rb_vos:
            self.rb_mgr.delete_role_binding_by_vo(rb_vo)

    def add_users_to_workspace_group(
        self,
        new_users_info_list: List[Dict[str, str]],
        new_users_role_map: Dict[str, str],
        workspace_group_workspace_ids: List[str],
        workspace_group_id: str,
        domain_id: str,
    ) -> List[Dict[str, str]]:
        workspace_group_new_users_info_list = []
        unique_user_ids = set()

        def add_user(user_info, workspace_group_workspace_id=None):
            if user_info["user_id"] not in unique_user_ids:
                role_type = new_users_role_map[user_info["role_id"]]
                user_data = {
                    "user_id": user_info["user_id"],
                    "role_id": user_info["role_id"],
                    "role_type": role_type,
                }

                if workspace_group_workspace_id:
                    role_binding_params = {
                        **user_data,
                        "resource_group": "WORKSPACE",
                        "domain_id": domain_id,
                        "workspace_group_id": workspace_group_id,
                        "workspace_id": workspace_group_workspace_id,
                    }
                    new_user_rb_vo = self.rb_svc.create_role_binding(
                        role_binding_params
                    )
                    user_data = {
                        "user_id": new_user_rb_vo.user_id,
                        "role_id": new_user_rb_vo.role_id,
                        "role_type": new_user_rb_vo.role_type,
                    }
                workspace_group_new_users_info_list.append(user_data)
                unique_user_ids.add(user_data["user_id"])

        if workspace_ids := workspace_group_workspace_ids:
            for workspace_id in workspace_ids:
                for new_user_info in new_users_info_list:
                    add_user(new_user_info, workspace_id)
        else:
            for new_user_info in new_users_info_list:
                add_user(new_user_info)

        return workspace_group_new_users_info_list

    def add_user_name_and_state_to_users(
        self,
        workspace_group_user_ids: List[str],
        workspace_group_info: Union[WorkspaceGroup, Dict[str, Any]],
        domain_id: str,
    ) -> Dict[str, Any]:
        """Add user's name and state to users in workspace group.
        Since the user's name and state are not in user of workspace group in database,
        we need to add user's name and state to users in the Application layer.
        Args:
            workspace_group_user_ids: 'List[str]'
            workspace_group_info: 'Union[WorkspaceGroup, Dict[str, Any]]'
            domain_id: 'str'
        Returns:
            workspace_group_info: 'Dict[str, Any]'
        """

        def update_user_info(
            user: Union[WorkspaceGroupUser, Dict[str, str]],
            user_info_dict: Dict[str, Dict[str, str]],
        ) -> Dict[str, str]:
            if isinstance(user, dict):
                user_id = user.get("user_id", "")
            else:
                user_id = getattr(user, "user_id", "") or ""
                user = user.to_mongo().to_dict()

            user_info = user_info_dict.get(user_id, {})

            user.update(
                {
                    "user_name": user_info.get("name", ""),
                    "state": user_info.get("state", ""),
                }
            )

            return user

        if isinstance(workspace_group_info, dict):
            workspace_group_users = workspace_group_info.get("users", [])
        else:
            workspace_group_users = workspace_group_info.users or []

        user_info_map = {}
        user_vos = self.user_mgr.filter_users(
            user_id=workspace_group_user_ids, domain_id=domain_id
        )
        for user_vo in user_vos:
            user_info_map[user_vo.user_id] = {
                "name": user_vo.name,
                "state": user_vo.state,
            }

        if workspace_group_users:
            updated_users = [
                update_user_info(user, user_info_map) for user in workspace_group_users
            ]
            if isinstance(workspace_group_info, dict):
                workspace_group_info["users"] = updated_users
            else:
                workspace_group_info.users = updated_users

        if not isinstance(workspace_group_info, dict):
            return workspace_group_info.to_dict()
        else:
            return workspace_group_info

    def get_users_info_list(
        self,
        new_users_info_list: List[Dict[str, str]],
        new_users_role_map: Dict[str, str],
        workspace_group_workspace_ids: List[str],
        workspace_group_vo: WorkspaceGroup,
        domain_id: str,
    ) -> List[Dict[str, str]]:
        workspace_group_old_users_info_list = workspace_group_vo.users or []
        workspace_group_new_users_info_list = self.add_users_to_workspace_group(
            new_users_info_list,
            new_users_role_map,
            workspace_group_workspace_ids,
            workspace_group_vo.workspace_group_id,
            domain_id,
        )
        return workspace_group_old_users_info_list + workspace_group_new_users_info_list

    def remove_users_from_workspace_group(
        self,
        user_ids: List[str],
        old_users: List[Dict[str, str]],
        workspace_group_id: str,
        domain_id: str,
        workspace_id: str = None,
    ) -> List[Dict[str, str]]:
        rb_vos = self.rb_mgr.filter_role_bindings(
            user_id=user_ids,
            workspace_group_id=workspace_group_id,
            domain_id=domain_id,
        )

        if rb_vos.count() > 0:
            for rb_vo in rb_vos:
                self.rb_mgr.delete_role_binding_by_vo(rb_vo)

        updated_users = [user for user in old_users if user["user_id"] not in user_ids]

        if not workspace_id:
            workspace_vos = self.workspace_mgr.filter_workspaces(
                workspace_group_id=workspace_group_id, domain_id=domain_id
            )
            for workspace_vo in workspace_vos:
                user_rb_ids = self.rb_mgr.stat_role_bindings(
                    query={
                        "distinct": "user_id",
                        "filter": [
                            {
                                "k": "workspace_id",
                                "v": workspace_vo.workspace_id,
                                "o": "eq",
                            },
                            {"k": "domain_id", "v": domain_id, "o": "eq"},
                        ],
                    }
                ).get("results", [])
                user_rb_total_count = len(user_rb_ids)

                self.workspace_mgr.update_workspace_by_vo(
                    {"user_count": user_rb_total_count}, workspace_vo
                )
        else:
            self.rb_svc.update_workspace_user_count(domain_id, workspace_id)

        return updated_users

    def get_workspace_groups_info(
        self, workspace_group_vos: QuerySet, domain_id
    ) -> List[Dict[str, Any]]:
        workspace_groups_info = []
        for workspace_group_vo in workspace_group_vos:
            workspace_group_user_ids = self._get_workspace_group_user_ids(
                workspace_group_vo
            )
            workspace_group_dict = self.add_user_name_and_state_to_users(
                workspace_group_user_ids, workspace_group_vo, domain_id
            )
            workspace_groups_info.append(workspace_group_dict)
        return workspace_groups_info

    @staticmethod
    def _get_workspace_group_user_ids(workspace_group_vo: WorkspaceGroup) -> List[str]:
        workspace_group_users = workspace_group_vo.users or []
        user_ids = set(user_info["user_id"] for user_info in workspace_group_users)
        return list(user_ids)

    @staticmethod
    def check_user_state(old_user_id: str, old_user_state: str) -> None:
        if old_user_state in ["DISABLED", "DELETED"]:
            _LOGGER.error(f"User ID {old_user_id}'s state is {old_user_state}.")
            raise ERROR_NOT_ALLOWED_USER_STATE(
                user_id=old_user_id, state=old_user_state
            )

    @staticmethod
    def check_role_type(role_type: str) -> None:
        if role_type not in ["WORKSPACE_OWNER", "WORKSPACE_MEMBER"]:
            raise ERROR_NOT_ALLOWED_ROLE_TYPE()

    def update_user_role_of_workspace_group(
        self,
        role_id: str,
        role_type: str,
        user_id: str,
        workspace_group_id: str,
        domain_id: str,
    ) -> None:
        role_binding_vos = self.rb_mgr.filter_role_bindings(
            user_id=user_id,
            workspace_group_id=workspace_group_id,
            domain_id=domain_id,
        )

        for role_binding_vo in role_binding_vos:
            role_binding_vo.update({"role_id": role_id, "role_type": role_type})



import logging
from typing import Union, List

from mongoengine import QuerySet
from spaceone.core.service import *
from spaceone.core.service.utils import *

from spaceone.identity.error.error_project_group import *
from spaceone.identity.manager.project_group_manager import ProjectGroupManager
from spaceone.identity.manager.resource_manager import ResourceManager
from spaceone.identity.manager.workspace_user_manager import WorkspaceUserManager
from spaceone.identity.model import ProjectGroup
from spaceone.identity.model.project_group.request import *
from spaceone.identity.model.project_group.response import *

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class ProjectGroupService(BaseService):
    resource = "ProjectGroup"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.project_group_mgr = ProjectGroupManager()
        self.resource_mgr = ResourceManager()

    @transaction(
        permission="identity:ProjectGroup.write", role_types=["WORKSPACE_OWNER"]
    )
    @convert_model
    def create(
        self, params: ProjectGroupCreateRequest
    ) -> Union[ProjectGroupResponse, dict]:
        """Create project group

        Args:
            params (ProjectGroupCreateRequest): {
                'name': 'str',              # required
                'tags': 'dict',
                'parent_group_id': 'str',
                'workspace_id': 'str',      # injected from auth (required)
                'domain_id': 'str'          # injected from auth (required)
            }
        Returns:
            ProjectGroupResponse:
        """

        if params.parent_group_id:
            self.project_group_mgr.get_project_group(
                params.parent_group_id,
                params.domain_id,
                params.workspace_id,
            )

        project_group_vo = self.project_group_mgr.create_project_group(params.dict())
        return ProjectGroupResponse(**project_group_vo.to_dict())

    @transaction(
        permission="identity:ProjectGroup.write", role_types=["WORKSPACE_OWNER"]
    )
    @convert_model
    def update(
        self, params: ProjectGroupUpdateRequest
    ) -> Union[ProjectGroupResponse, dict]:
        """Update project group

        Args:
            params (ProjectGroupUpdateRequest): {
                'project_group_id': 'str',      # required
                'name': 'str',
                'tags': 'dict',
                'domain_id': 'str',             # injected from auth (required)
                'workspace_id': 'str',          # injected from auth (required)
            }
            Returns:
                ProjectGroupResponse:
        """

        project_group_vo = self.project_group_mgr.get_project_group(
            params.project_group_id,
            params.domain_id,
            params.workspace_id,
        )

        # Check is managed resource
        self.resource_mgr.check_is_managed_resource_by_trusted_account(project_group_vo)

        project_group_vo = self.project_group_mgr.update_project_group_by_vo(
            params.dict(exclude_unset=True), project_group_vo
        )

        return ProjectGroupResponse(**project_group_vo.to_dict())

    @transaction(
        permission="identity:ProjectGroup.write", role_types=["WORKSPACE_OWNER"]
    )
    @convert_model
    def change_parent_group(
        self, params: ProjectChangeParentGroupRequest
    ) -> Union[ProjectGroupResponse, dict]:
        """Change parent project group

        Args:
            params (ProjectChangeParentGroupRequest): {
                'project_group_id': 'str',      # required
                'parent_group_id': 'str',       # required
                'workspace_id': 'str',          # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }
            Returns:
                ProjectGroupResponse:
        """

        project_group_vo = self.project_group_mgr.get_project_group(
            params.project_group_id,
            params.domain_id,
            params.workspace_id,
        )

        # Check is managed resource
        self.resource_mgr.check_is_managed_resource_by_trusted_account(project_group_vo)

        # Check parent project group is
        if params.parent_group_id:
            self.project_group_mgr.get_project_group(
                params.parent_group_id, params.domain_id, params.workspace_id
            )

            # Check parent project group is not sub project group
            project_group_vos = self.project_group_mgr.filter_project_groups(
                workspace_id=params.workspace_id, domain_id=params.domain_id
            )

            self._check_is_sub_project_group(
                params.parent_group_id,
                project_group_vo.project_group_id,
                project_group_vos,
            )

        project_group_vo = self.project_group_mgr.update_project_group_by_vo(
            params.dict(), project_group_vo
        )

        return ProjectGroupResponse(**project_group_vo.to_dict())

    @transaction(
        permission="identity:ProjectGroup.write", role_types=["WORKSPACE_OWNER"]
    )
    @convert_model
    def delete(self, params: ProjectGroupDeleteRequest) -> None:
        """Delete project group

        Args:
            params (ProjectGroupDeleteRequest): {
                'project_group_id': 'str',      # required
                'workspace_id': 'str',          # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }
        Returns:
            None
        """

        project_group_vo = self.project_group_mgr.get_project_group(
            params.project_group_id,
            params.domain_id,
            params.workspace_id,
        )

        # Check is managed resource
        self.resource_mgr.check_is_managed_resource_by_trusted_account(project_group_vo)

        self.project_group_mgr.delete_project_group_by_vo(project_group_vo)

    @transaction(
        permission="identity:ProjectGroup.write", role_types=["WORKSPACE_OWNER"]
    )
    @convert_model
    def add_users(
        self, params: ProjectGroupAddUsersRequest
    ) -> Union[ProjectGroupResponse, dict]:
        """Add users to project group

        Args:
            params (ProjectGroupAddUsersRequest): {
                'project_group_id': 'str',      # required
                'users': 'list',             # required
                'domain_id': 'str',             # injected from auth (required)
                'workspace_id': 'str'           # injected from auth (required)
            }
        Returns:
            ProjectGroupResponse:
        """

        project_group_vo = self.project_group_mgr.get_project_group(
            params.project_group_id, params.domain_id, params.workspace_id
        )

        if len(params.users) > 0:
            workspace_user_mgr = WorkspaceUserManager()
            workspace_user_mgr.check_workspace_users(
                params.users, params.workspace_id, params.domain_id
            )

            users = project_group_vo.users or []
            users.extend(params.users)
            params.users = list(set(users))

            project_group_vo = self.project_group_mgr.update_project_group_by_vo(
                params.dict(exclude_unset=True), project_group_vo
            )

        return ProjectGroupResponse(**project_group_vo.to_dict())

    @transaction(
        permission="identity:ProjectGroup.write",
        role_types=["WORKSPACE_OWNER"],
    )
    @convert_model
    def remove_users(
        self, params: ProjectGroupRemoveUsersRequest
    ) -> ProjectGroupResponse:
        """Remove users from project group
        Args:
            params: (ProjectGroupRemoveUsersRequest): {
                'project_group_id': 'str',      # required
                'users': 'list',             # required
                'workspace_id': 'str',             # injected from auth (required)
                'domain_id': 'str'           # injected from auth (required)
            }

        """

        project_group_vo = self.project_group_mgr.get_project_group(
            params.project_group_id, params.domain_id, params.workspace_id
        )

        if len(params.users) > 0:
            workspace_user_mgr = WorkspaceUserManager()
            workspace_user_mgr.check_workspace_users(
                params.users, params.workspace_id, params.domain_id
            )

            users = project_group_vo.users or []
            params.users = list(set(users) - set(params.users))
            project_group_vo = self.project_group_mgr.update_project_group_by_vo(
                params.dict(exclude_unset=True), project_group_vo
            )

        return ProjectGroupResponse(**project_group_vo.to_dict())

    @transaction(
        permission="identity:ProjectGroup.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def get(self, params: ProjectGroupGetRequest) -> Union[ProjectGroupResponse, dict]:
        """Get project group

        Args:
            params (ProjectGroupGetRequest): {
                'project_group_id': 'str',      # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
            }
        Returns:
            ProjectGroupResponse:
        """

        project_group_vo = self.project_group_mgr.get_project_group(
            params.project_group_id,
            params.domain_id,
            params.workspace_id,
        )

        return ProjectGroupResponse(**project_group_vo.to_dict())

    @transaction(
        permission="identity:ProjectGroup.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @append_query_filter(
        ["project_group_id", "name", "parent_group_id", "workspace_id", "domain_id"]
    )
    @append_keyword_filter(["project_group_id", "name"])
    @convert_model
    def list(
        self, params: ProjectGroupSearchQueryRequest
    ) -> Union[ProjectGroupsResponse, dict]:
        """List project groups

        Args:
            params (ProjectGroupSearchQueryRequest): {
                'query': 'dict (spaceone.api.core.v1.Query)',
                'project_group_id': 'str',
                'name': 'str',
                'parent_group_id': 'str',
                'workspace_id': 'str',      # injected from auth
                'domain_id': 'str',         # injected from auth (required)
            }
        Returns:
            ProjectGroupsResponse:
        """

        query = params.query or {}
        project_vos, total_count = self.project_group_mgr.list_project_groups(query)

        projects_info = [project_vo.to_dict() for project_vo in project_vos]
        return ProjectGroupsResponse(results=projects_info, total_count=total_count)

    @transaction(
        permission="identity:ProjectGroup.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @append_query_filter(["workspace_id", "domain_id"])
    @append_keyword_filter(["project_group_id", "name"])
    @convert_model
    def stat(self, params: ProjectGroupStatQueryRequest) -> dict:
        """Stat project groups
        Args:
            params (ProjectGroupStatQueryRequest): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)', # required
                'workspace_id': 'str',    # injected from auth
                'domain_id': 'str',       # injected from auth (required)
            }
        Returns:
            dict: {
                'results': 'list',
                'total_count': 'int'
            }
        """

        query = params.query or {}
        return self.project_group_mgr.stat_project_groups(query)

    def _check_is_sub_project_group(
        self,
        change_parent_group_id: str,
        cur_group_id: str,
        project_group_vos: QuerySet,
    ) -> Union[None, Exception]:
        for project_group_vo in project_group_vos:
            if project_group_vo.parent_group_id == cur_group_id:
                if change_parent_group_id == project_group_vo.project_group_id:
                    raise ERROR_NOT_ALLOWED_TO_CHANGE_PARENT_GROUP_TO_SUB_PROJECT_GROUP(
                        project_group_id=change_parent_group_id
                    )
                else:
                    return self._check_is_sub_project_group(
                        change_parent_group_id,
                        project_group_vo.project_group_id,
                        project_group_vos,
                    )
        return None

    def _check_workspace_member_permission(
        self, project_group_vo: ProjectGroup
    ) -> None:
        role_type = self.transaction.get_meta("authorization.role_type")
        if role_type == "WORKSPACE_MEMBER":
            user_id = self.transaction.get_meta("authorization.user_id")
            project_group_users = project_group_vo.users or []
            if user_id not in project_group_users:
                raise ERROR_USER_NOT_IN_PROJECT_GROUP(
                    user_id=user_id, project_group_id=project_group_vo.project_group_id
                )



from typing import Union, Literal
from pydantic import BaseModel, Field

__all__ = [
    "ExternalAuthSetRequest",
    "ExternalAuthUnsetRequest",
    "ExternalAuthGetRequest",
]

UpgradeMode = Literal["AUTO", "MANUAL"]


class Plugin(BaseModel):
    plugin_id: str
    version: Union[str, None] = None
    upgrade_mode: Union[str, None] = None
    options: Union[dict, None] = None
    secret_data: Union[dict, None] = None
    schema_id: Union[str, None] = Field(None, alias="schema")


class ExternalAuthSetRequest(BaseModel):
    domain_id: str
    plugin_info: Plugin


class ExternalAuthUnsetRequest(BaseModel):
    domain_id: str


class ExternalAuthGetRequest(BaseModel):
    domain_id: str



import logging
from typing import Tuple, List

from spaceone.core import cache
from spaceone.core.manager import BaseManager
from spaceone.identity.model.provider.database import Provider
from spaceone.identity.manager.managed_resource_manager import ManagedResourceManager

_LOGGER = logging.getLogger(__name__)


class ProviderManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.provider_model = Provider

    def create_provider(self, params: dict) -> Provider:
        def _rollback(vo: Provider):
            _LOGGER.info(f"[create_provider._rollback] Delete provider : {vo.provider}")
            vo.delete()

        provider_vo = self.provider_model.create(params)
        self.transaction.add_rollback(_rollback, provider_vo)

        return provider_vo

    def update_provider_by_vo(self, params: dict, provider_vo: Provider) -> Provider:
        def _rollback(old_data):
            _LOGGER.info(
                f'[update_provider._rollback] Revert Data : {old_data["provider"]}'
            )
            provider_vo.update(old_data)

        self.transaction.add_rollback(_rollback, provider_vo.to_dict())

        return provider_vo.update(params)

    @staticmethod
    def delete_provider_by_vo(provider_vo: Provider) -> None:
        provider_vo.delete()

    def get_provider(self, provider: str, domain_id: str) -> Provider:
        return self.provider_model.get(provider=provider, domain_id=domain_id)

    def filter_providers(self, **conditions) -> List[Provider]:
        return self.provider_model.filter(**conditions)

    def list_providers(self, query: dict, domain_id: str) -> Tuple[list, int]:
        self._create_managed_provider(domain_id)
        return self.provider_model.query(**query)

    def stat_providers(self, query: dict) -> dict:
        return self.provider_model.stat(**query)

    @cache.cacheable(key="identity:managed-provider:{domain_id}:sync", expire=300)
    def _create_managed_provider(self, domain_id: str) -> bool:
        managed_resource_mgr = ManagedResourceManager()

        provider_vos = self.filter_providers(domain_id=domain_id, is_managed=True)

        installed_provider_version_map = {}
        for provider_vo in provider_vos:
            installed_provider_version_map[provider_vo.provider] = provider_vo.version

        managed_provider_map = managed_resource_mgr.get_managed_providers()

        for managed_provider, managed_provider_info in managed_provider_map.items():
            managed_provider_info["domain_id"] = domain_id
            managed_provider_info["is_managed"] = True

            if provider_version := installed_provider_version_map.get(managed_provider):
                if provider_version != managed_provider_info["version"]:
                    _LOGGER.debug(
                        f"[_create_managed_provider] update managed provider: {managed_provider}"
                    )
                    provider_vo = self.get_provider(managed_provider, domain_id)
                    self.update_provider_by_vo(managed_provider_info, provider_vo)
            else:
                _LOGGER.debug(
                    f"[_create_managed_provider] create new managed provider: {managed_provider}"
                )
                self.create_provider(managed_provider_info)

        return True



from spaceone.core.error import *


class ERROR_GENERATE_KEY_FAILURE(ERROR_BASE):
    _message = "Error on generate key."


class ERROR_MANAGED_RESOURCE_CAN_NOT_BE_MODIFIED(ERROR_UNKNOWN):
    message = "Managed resource can not be deleted. please disable schedule first."


class ERROR_WORKSPACES_DO_NOT_EXIST(ERROR_UNKNOWN):
    _message = "Resource Not Found. (key = {key}, reason = {reason})"


class ERROR_ROLE_DOES_NOT_EXIST_OF_USER(ERROR_NOT_FOUND):
    _message = "Role does not exist in User. (role_id = {role_id}, user_id = {user_id})"


class ERROR_NOT_ALLOWED_TO_DELETE_ROLE_BINDING(ERROR_INVALID_ARGUMENT):
    _message = "Not allowed to delete role binding. (workspace_group_id = {workspace_group_id}, role_binding_id = {role_binding_id})"


class ERROR_ROLE_IN_USED_AT_ROLE_BINDING(ERROR_INVALID_ARGUMENT):
    _message = "Role is in used at RoleBinding. (role_id = {role_id})"


class ERROR_WORKSPACE_EXIST_IN_WORKSPACE_GROUP(ERROR_INVALID_ARGUMENT):
    _message = """Workspace exists in Workspace Group. (workspace_id = {workspace_id}, workspace_group_id = {workspace_group_id})
               Remove the workspace from the workspace group before deleting the workspace group."""


class ERROR_USER_EXIST_IN_WORKSPACE_GROUP(ERROR_INVALID_ARGUMENT):
    _message = """User exists in Workspace Group. (user_id = {user_id}, workspace_group_id = {workspace_group_id})
               Remove the user from the workspace group before deleting the workspace group."""
